# -*- coding: utf-8 -*-

import os.path
from pathlib import Path
import numpy as np
import pandas as pd
from numpy.testing import assert_equal, assert_array_equal
from numpy.testing import assert_allclose
import scipy.sparse
import itertools
import pytest
from sklearn.preprocessing import normalize

from freediscovery.engine.vectorizer import (FeatureVectorizer,
                                             _FeatureVectorizerSampled)
from freediscovery.engine.ingestion import DocumentIndex

from freediscovery.tests.run_suite import check_cache
from freediscovery.exceptions import (NotFound, WrongParameter)

basename = os.path.dirname(__file__)
data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")
n_features = 1100000


@pytest.mark.parametrize("analyzer, ngram_range, use_hashing",
                         list(itertools.product(['word', 'char'],
                                                [[1, 1], [1, 2]],
                                                ['hashed', 'non-hashed'])))
def test_feature_extraction_tokenization(analyzer, ngram_range, use_hashing):
    cache_dir = check_cache()
    use_hashing = (use_hashing == 'hashed')

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(analyzer=analyzer, ngram_range=ngram_range,
                    use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    res2 = fe._load_features(uuid)
    assert isinstance(res2,  np.ndarray) or scipy.sparse.issparse(res2), "not an array {}".format(res2)

    assert np.isfinite(res2.data).all()

    assert_allclose(normalize(res2).data, res2.data)  # data is l2 normalized

    fe.delete()


def test_feature_extraction_storage():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir, file_pattern='.*\d.txt')
    db = pd.read_pickle(os.path.join(cache_dir, 'ediscovery_cache',
                                     uuid, 'db'))
    assert 'file_path' not in db.columns


@pytest.mark.parametrize("weighting, use_hashing",
                         list(itertools.product(("".join(el) for el in itertools.product('nlb', 'ns', 'c')),
                                                ['hashed', ''])))
def test_feature_extraction_weighting(weighting,
                                      use_hashing):
    cache_dir = check_cache()

    use_hashing = (use_hashing == 'hashed')

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(weighting=weighting, use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    res2 = fe._load_features(uuid)
    assert isinstance(res2,  np.ndarray) or scipy.sparse.issparse(res2), \
        "not an array {}".format(res2)

    assert np.isfinite(res2.data).all()
    assert_allclose(normalize(res2).data, res2.data)  # data is l2 normalized

    fe.delete()


@pytest.mark.parametrize("n_features, weighting, use_hashing",
                         list(itertools.product([None, 4, 1000],
                                                ['nnc', 'nsc'],
                                                ['hashed', ''])))
def test_feature_extraction_nfeatures(n_features, weighting, use_hashing):
    cache_dir = check_cache()

    use_hashing = (use_hashing == 'hashed')

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(n_features=n_features, weighting=weighting, use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    res2 = fe._load_features(uuid)
    assert isinstance(res2,  np.ndarray) or scipy.sparse.issparse(res2), \
        "not an array {}".format(res2)

    assert np.isfinite(res2.data).all()

    assert res2.shape[1] == fe.n_features_

    fe.delete()


@pytest.mark.parametrize('use_hashing,', [True, False])
def test_search_filenames(use_hashing):
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    assert fe.db_ is not None

    for low, high, step in [(0, 1, 1),
                            (0, 4, 1),
                            (3, 1, -1)]:
        idx_slice = list(range(low, high, step))
        filenames_slice = [fe.filenames_[idx] for idx in idx_slice]
        idx0 = fe.db_._search_filenames(filenames_slice)
        assert_equal(idx0, idx_slice)
        assert_equal(filenames_slice, fe[idx0])

    with pytest.raises(NotFound):
        fe.db_._search_filenames(['DOES_NOT_EXIST.txt'])

    if not use_hashing:
        n_top_words = 5
        terms = fe.query_features([2, 3, 5], n_top_words=n_top_words)
        assert len(terms) == n_top_words

    fe.list_datasets()


@pytest.mark.parametrize('use_hashing, min_df, max_df', [[False, 0.1, 0.6],
                                                         [True,  0.1, 0.6]])
def test_df_filtering(use_hashing, min_df, max_df):
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(min_df=min_df, max_df=max_df, use_hashing=use_hashing)
    fe.ingest(data_dir)

    X = fe._load_features(uuid)

    fe2 = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid2 = fe2.setup(use_hashing=use_hashing)
    fe2.ingest(data_dir)

    X2 = fe2._load_features(uuid2)

    if use_hashing:
        # min/max_df does not affect the number of features
        assert X.shape[1] == X2.shape[1]
    else:
        # min/max_df removes some features
        assert X.shape[1] < X2.shape[1]

    fe.delete()


def test_append_documents():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    X = fe._load_features(uuid)
    db = fe.db_
    filenames = fe.filenames_
    n_samples = len(fe.filenames_)

    docs = DocumentIndex.from_folder(data_dir).data
    docs['document_id'] += 10
    dataset_definition = docs[['file_path', 'document_id']].to_dict(orient='records')
    for row in dataset_definition:
        row['file_path'] = os.path.join(data_dir, row['file_path'])
    fe.append(dataset_definition)
    X_new = fe._load_features(uuid)
    assert X_new.shape[0] == X.shape[0]*2
    assert fe.db_.data.shape[0] == db.data.shape[0]*2
    assert len(fe.filenames_) == len(filenames)*2

    dbn = fe.db_.data
    assert_equal(dbn.iloc[:n_samples]['document_id'].values,
                 dbn.iloc[n_samples:]['document_id'].values - 10)
    # check that internal id is contiguous
    assert (np.diff(dbn.internal_id.values) == 1).all()

    # check the number of samples is consistent
    del fe._pars
    assert fe.n_samples_ == n_samples * 2

    fe.delete()


def test_remove_documents():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    X = fe._load_features(uuid)
    db = fe.db_.data
    filenames = fe.filenames_
    n_samples = len(fe.filenames_)

    docs = DocumentIndex.from_folder(data_dir).data
    dataset_definition = docs[['document_id']].to_dict(orient='records')
    fe.remove([dataset_definition[2], dataset_definition[4]])
    X_new = fe._load_features(uuid)
    assert X_new.shape[0] == X.shape[0] - 2
    assert fe.db_.data.shape[0] == db.shape[0] - 2
    assert len(fe.filenames_) == len(filenames) - 2

    dbn = fe.db_.data
    assert_equal(db.iloc[[0, 1, 3, 5]]['document_id'].values,
                 dbn['document_id'].values)
    # check that internal id is contiguous
    assert (np.diff(dbn.internal_id.values) == 1).all()

    # check the number of samples is consistent
    del fe._pars
    assert fe.n_samples_ == n_samples - 2

    fe.delete()


def test_sampling_filenames():
    cache_dir = check_cache()

    fe_pars = {'weighting': 'bnn'}

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    with pytest.warns(UserWarning):
        # there is a warning because we don't use norm='l2'
        uuid = fe.setup(use_hashing=True, **fe_pars)
        fe.ingest(data_dir, file_pattern='.*\d.txt')
    X = fe._load_features(uuid)

    # don't use any sampling
    fes = _FeatureVectorizerSampled(cache_dir=cache_dir, dsid=uuid,
                                    sampling_filenames=None)
    X_s = fes._load_features(uuid)
    pars = fe.pars_
    fnames = fe.filenames_
    fnames_s = fes.filenames_
    assert_array_equal(fnames, fnames_s)
    assert_array_equal(X.data, X_s.data)
    assert fes.n_samples_ == len(fnames)

    fes = _FeatureVectorizerSampled(cache_dir=cache_dir, dsid=uuid,
                                    sampling_filenames=fnames[::-1])

    assert fes.sampling_index is not None
    X_s = fes._load_features(uuid)
    pars_s = fes.pars_
    fnames_s = fes.filenames_
    assert_array_equal(fnames[::-1], fnames_s)
    assert_array_equal(X[::-1, :].data, X_s.data)
    for key in pars:
        if key == 'filenames':
            assert pars[key][::-1] == pars_s[key]
        else:
            assert pars[key] == pars_s[key]

    # repeat twice the filenames
    fes = _FeatureVectorizerSampled(cache_dir=cache_dir, dsid=uuid,
                                    sampling_filenames=(fnames+fnames))

    assert fes.sampling_index is not None
    X_s = fes._load_features(uuid)
    pars_s = fes.pars_
    fnames_s = fes.filenames_
    assert_array_equal(fnames + fnames, fnames_s)
    assert_array_equal(X.data, X_s[:len(fnames)].data)
    assert_array_equal(X.data, X_s[len(fnames):].data)
    assert fes.n_samples_ == len(fnames)*2
    # for key in pars:
    #    assert pars[key] == pars_s[key]

    # downsample the filenames
    N = len(fnames)//2

    np.random.seed(1)

    idx = np.random.choice(fe.n_samples_, size=(N,))
    fnames_s_in = np.array(fnames)[idx].tolist()

    fes = _FeatureVectorizerSampled(cache_dir=cache_dir, dsid=uuid,
                                    sampling_filenames=fnames_s_in)

    assert fes.sampling_index is not None
    X_s = fes._load_features(uuid)
    pars_s = fes.pars_
    fnames_s = fes.filenames_
    assert_array_equal(fnames_s_in, fnames_s)
    assert_array_equal(X[idx].data, X_s.data)
    assert fes.n_samples_ == N

    fe.delete()


@pytest.mark.parametrize("use_hashing", ['hashed', 'non-hashed'])
def test_feature_extraction_cyrillic(use_hashing):
    data_dir = os.path.join(basename, "..", "..", "data", "ds_002", "raw")
    cache_dir = check_cache()
    use_hashing = (use_hashing == 'hashed')

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup(use_hashing=use_hashing)
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    res2 = fe._load_features(uuid)

    filenames = fe.filenames_
    fe._filenames = None
    filenames2 = fe.filenames_

    assert_equal(filenames2, filenames)
    assert isinstance(res2,  np.ndarray) or scipy.sparse.issparse(res2),\
        "not an array {}".format(res2)

    assert np.isfinite(res2.data).all()
    fe.delete()


def test_email_parsing():
    data_dir = os.path.join(basename, "..", "..", "data",
                            "fedora-devel-list-2008-October")
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir)

    email_md = fe.parse_email_headers()
    assert len(fe.filenames_) == len(email_md)

    fe.delete()


def test_ingestion_batches():
    data_dir = os.path.join(basename, "..", "..", "data", "ds_002", "raw")
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    with pytest.raises(ValueError):
        fe.ingest(vectorize=True)  # no ingested files
    fe.ingest(data_dir, file_pattern='.*\d.txt', vectorize=False)
    fe.ingest(data_dir, file_pattern='.*\d.txt', vectorize=False)
    fe.ingest(data_dir, file_pattern='.*\d.txt', vectorize=False)

    fe.ingest(vectorize=True)

    assert fe.db_.data.shape[0] == len(fe.filenames_)
    assert len(fe.filenames_) == 6*3
    X = fe._load_features()
    assert X.shape[0] == 6*3

    with pytest.raises(ValueError):
        fe.ingest(vectorize=True)  # already vectorized


def test_ingestion_content():
    data_dir = Path(basename, "..", "..", "data", "ds_002", "raw")

    dd = []
    for idx, fname in enumerate(sorted(data_dir.glob('*txt'))):
        with fname.open('rt', encoding='utf-8') as fh:
            dd.append({'document_id': idx + 19,
                       'content': fh.read()})
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(dataset_definition=dd, vectorize=True)
    assert len(fe.filenames_) == 6
    assert fe.filenames_[0] == '000000000_0.txt'
    X = fe._load_features()
    assert X.shape[0] == 6
    assert fe.db_.data.shape[0] == len(fe.filenames_)

    fe2 = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    fe2.setup()
    fe2.ingest(data_dir=str(data_dir))

    X2 = fe2._load_features()
    assert X.shape == X2.shape
    assert_array_equal(X.indices, X2.indices)
    assert_array_equal(X.data, X2.data)

def test_non_random_dsid():
    data_dir = os.path.join(basename, "..", "..", "data", "ds_002", "raw")
    cache_dir = check_cache()

    dsid = 'test-dataset'

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w', dsid=dsid)
    uuid = fe.setup()
    assert dsid == uuid
    fe.ingest(data_dir, file_pattern='.*\d.txt', vectorize=False)
    # writing with the same name fails
    with pytest.raises(WrongParameter):
        FeatureVectorizer(cache_dir=cache_dir, mode='w', dsid=dsid)

    FeatureVectorizer(cache_dir=cache_dir, mode='r', dsid=dsid)

    FeatureVectorizer(cache_dir=cache_dir, mode='fw', dsid=dsid)
    # special characters are not allowed
    with pytest.raises(WrongParameter):
        fh = FeatureVectorizer(cache_dir=cache_dir, mode='fw', dsid='?+ds$$')
        uuid = fh.setup()



# -*- coding: utf-8 -*-

import pytest
import json
import os.path

import numpy as np

from .. import fd_app
from ...tests.run_suite import check_cache
from ...utils import dict2type

from sklearn.externals.joblib import Memory

V01 = '/api/v0'

data_dir = os.path.dirname(__file__)
email_data_dir = os.path.join(data_dir, "..", "..", "data", "fedora-devel-list-2008-October")
data_dir = os.path.join(data_dir, "..", "..", "data", "ds_001", "raw")
CACHE_DIR = check_cache()

np.random.seed(43)


def parse_res(res):
    return json.loads(res.data.decode('utf-8'))


def app_call_wrapper(func):
    """Wrapp the POST, GET, DELETE methods of flask.testing.FlaskClient
    in order to make the response_code checks and convert the
    json output to dict that we always need to do
    """

    def inner_function(*args, valid_return_codes=None, **kwargs):
        if valid_return_codes is None:
            valid_return_codes = [200]
        res = func(*args, **kwargs)
        assert res.status_code in valid_return_codes, args[0]
        data = parse_res(res)
        return data
    return inner_function


@pytest.fixture
def app():
    tapp = fd_app(CACHE_DIR)
    tapp.config['TESTING'] = True

    client = tapp.test_client()
    client.post_check = app_call_wrapper(client.post)
    client.get_check = app_call_wrapper(client.get)
    client.delete_check = app_call_wrapper(client.delete)
    return client


@pytest.fixture
def app_notest():
    tapp = fd_app(CACHE_DIR)
    tapp.config['TESTING'] = False

    client = tapp.test_client()
    client.post_check = app_call_wrapper(client.post)
    client.get_check = app_call_wrapper(client.get)
    client.delete_check = app_call_wrapper(client.delete)
    return client


memory = Memory(cachedir=os.path.join(CACHE_DIR, '_joblib_cache'), verbose=0)

# ===========================================================================#
#
#                     Feature extraction
#
# ===========================================================================#


def get_features(app, hashed=False, metadata_fields='data_dir',
                 n_categories=2, dataset='20_newsgroups_3categories', **kwargs):

    method = V01 + "/feature-extraction/"
    pars = {'use_hashing': hashed}
    pars.update(kwargs)
    data = app.post_check(method, json=pars)

    assert dict2type(data, collapse_lists=True) == {'id': 'str'}
    dsid = data['id']

    pars = {}
    if not kwargs.get('parse_email_headers'):
        url = V01 + '/example-dataset/{}'.format(dataset)
        res = app.get(url, json={'n_categories': n_categories})
        assert res.status_code == 200, url
        input_ds = parse_res(res)
        data_dir = input_ds['metadata']['data_dir']
        pars['dataset_definition'] = [{'document_id': row['document_id'],
                                       'file_path': os.path.join(data_dir,
                                                                 row['file_path'])}
                                      for row in input_ds['dataset']]
    else:
        pars['data_dir'] = email_data_dir
        input_ds = None

    method = V01 + "/feature-extraction/{}".format(dsid)
    res = app.post_check(method, json=pars)
    assert dict2type(res) == {'id': 'str'}
    pars = app.get_check(method)
    pars.pop('filenames')
    pars.pop('dataset_definition', None)
    return dsid, pars, input_ds


@memory.cache(ignore=['app'])
def get_features_cached(app, hashed=False, n_categories=2,
                        dataset='20_newsgroups_3categories', **kwargs):
    return get_features(app, hashed=hashed, n_categories=n_categories,
                        dataset=dataset, **kwargs)


def get_features_lsi(app, hashed=False, metadata_fields='data_dir', **kwargs):
    dsid, pars = get_features(app, hashed=hashed,
                              metadata_fields=metadata_fields, **kwargs)
    lsi_pars = dict(n_components=201, parent_id=dsid)
    method = V01 + "/lsi/"
    res = app.post(method, json=lsi_pars)
    assert res.status_code == 200
    data = parse_res(res)
    assert dict2type(data) == {'explained_variance': 'float', 'id': 'str'}
    lsi_id = data['id']
    return dsid, lsi_id, pars


@memory.cache(ignore=['app'])
def get_features_lsi_cached(app, hashed=False, n_categories=2, n_components=201,
                            dataset="20_newsgroups_3categories"):
    dsid, pars, input_ds = get_features_cached(app, hashed=hashed,
                                               n_categories=n_categories)
    lsi_pars = dict(n_components=n_components, parent_id=dsid)
    method = V01 + "/lsi/"
    res = app.post(method, json=lsi_pars)
    assert res.status_code == 200
    data = parse_res(res)
    assert dict2type(data) == {'explained_variance': 'float', 'id': 'str'}
    lsi_id = data['id']
    return dsid, lsi_id, pars, input_ds



# Authors: Roman Yurchak
#
# License: BSD 3 clause


class _BaseException(Exception):
    status_code = -1
    message = 'base exception'

    def __init__(self, message=None):
        Exception.__init__(self, message)
        if message is not None:
            self.message = message

    def to_dict(self):
        rv = {'message':  '{} {}: {}'.format(self.status_code,
              type(self).__name__, self.message)}
        return rv


class NotFound(_BaseException):
    status_code = 404
    message = 'Not Found'


class DatasetNotFound(NotFound):
    status_code = 500
    message = 'Dataset Not Found'


class ModelNotFound(NotFound):
    status_code = 500
    message = 'Model Not Found'


class InitException(_BaseException):
    status_code = 500
    message = 'Model Not Found'


class WrongParameter(_BaseException):
    status_code = 500
    message = 'Model Not Found'


class NotImplementedFD(_BaseException):
    status_code = 500
    message = 'Not implemented in FreeDiscovery'


class OptionalDependencyMissing(_BaseException):
    status_code = 500
    message = 'Optional dependency is missing'



# -*- coding: utf-8 -*-

import os.path
from pathlib import Path
import pickle

import pytest
import numpy as np
import pandas as pd
from numpy.testing import assert_equal, assert_array_equal

from freediscovery.utils import dict2type, sdict_keys
from .base import (parse_res, V01, app, data_dir,
                   CACHE_DIR)
from sklearn.externals import joblib


@pytest.mark.parametrize('preprocess', [[],
                                           ['emails_ignore_header']])
def test_preprocessing_email_headers(app, preprocess):
    method = V01 + "/feature-extraction/"
    data = app.post_check(method, json={"preprocess": preprocess})
    dsid = data['id']
    method += dsid
    app.post_check(method, json={'dataset_definition': [{'file_path': os.path.join(data_dir, '0.7.6.28635.txt')}]})


    with (Path(CACHE_DIR) / 'ediscovery_cache' / dsid / 'vectorizer').open('rb') as fh:
        vectorizer = pickle.load(fh)


    pars = app.get_check(method)
    assert pars['preprocess'] == preprocess


    vocabulary = list(vectorizer.vocabulary_.keys())
    if not preprocess:
        assert 'baumbach' in vocabulary # a name in the To: header
    else:
        assert 'baumbach' not in vocabulary # a name in the To: header



# -*- coding: utf-8 -*-

from .base import (select_top_words, ClusterLabels)
from .utils import centroid_similarity
from .birch import Birch
from .optimal_sampling import compute_optimal_sampling

from .dendrogram import _DendrogramChildren
from .hierarchy import birch_hierarchy_wrapper, BirchSubcluster



# -*- coding: utf-8 -*-

import numpy as np
from scipy.special import logit, expit
from sklearn.externals import joblib
from sklearn.preprocessing import LabelEncoder

from freediscovery.engine.base import _BaseWrapper
from freediscovery.utils import setup_model, _rename_main_thread
from freediscovery.neighbors import NearestCentroidRanker, NearestNeighborRanker
from freediscovery.metrics import _scale_cosine_similarity
from freediscovery.exceptions import (WrongParameter, NotImplementedFD,
                                      OptionalDependencyMissing)


class _CategorizerWrapper(_BaseWrapper):
    """ Document categorization model

    The option `use_hashing=True` must be set for the feature extraction.
    Recommended options also include, `use_idf=1, sublinear_tf=0, binary=0`.

    Parameters
    ----------
    cache_dir : str
      folder where the model will be saved
    parent_id : str, optional
      dataset id
    mid : str, optional
      model id
    cv_scoring : str, optional, default='roc_auc'
      score that is used for Cross Validation, cf. sklearn
    cv_n_folds : str, optional
      number of K-folds used for Cross Validation
    """

    _wrapper_type = "categorizer"

    def __init__(self, cache_dir='/tmp/',  parent_id=None, mid=None,
                 cv_scoring='roc_auc', cv_n_folds=3, random_state=None):

        super(_CategorizerWrapper, self).__init__(cache_dir=cache_dir,
                                          parent_id=parent_id,
                                          mid=mid, load_model=True)

        if mid is not None:
            self.le = joblib.load(str(self.model_dir / mid / 'label_encoder'))
        self.cv_scoring = cv_scoring
        self.cv_n_folds = cv_n_folds
        self.random_state = random_state

    @staticmethod
    def _build_estimator(Y_train, method, cv, cv_scoring, cv_n_folds, random_state=None, **options):
        if cv:
            #from sklearn.cross_validation import StratifiedKFold
            #cv_obj = StratifiedKFold(n_splits=cv_n_folds, shuffle=False)
            cv_obj = cv_n_folds  # temporary hack (due to piclking issues otherwise, this needs to be fixed)
        else:
            cv_obj = None

        _rename_main_thread()

        if method == 'LinearSVC':
            from sklearn.svm import LinearSVC
            if cv is None:
                cmod = LinearSVC(random_state=random_state, **options)
            else:
                try:
                    from freediscovery_extra import make_linearsvc_cv_model
                except ImportError:
                    raise OptionalDependencyMissing('freediscovery_extra')
                cmod = make_linearsvc_cv_model(cv_obj, cv_scoring, **options)
        elif method == 'LogisticRegression':
            from sklearn.linear_model import LogisticRegression
            if cv is None:
                cmod = LogisticRegression(random_state=random_state, **options)
            else:
                try:
                    from freediscovery_extra import make_logregr_cv_model
                except ImportError:
                    raise OptionalDependencyMissing('freediscovery_extra')
                cmod = make_logregr_cv_model(cv_obj, cv_scoring, **options)
        elif method == 'NearestCentroid':
            cmod = NearestCentroidRanker()
        elif method == 'NearestNeighbor':
            cmod = NearestNeighborRanker()
        elif method == 'xgboost':
            try:
                import xgboost as xgb
            except ImportError:
                raise OptionalDependencyMissing('xgboost')
            if cv is None:
                try:
                    from freediscovery_extra import make_xgboost_model
                except ImportError:
                    raise OptionalDependencyMissing('freediscovery_extra')
                cmod = make_xgboost_model(cv_obj, cv_scoring, **options)
            else:
                try:
                    from freediscovery_extra import make_xgboost_cv_model
                except ImportError:
                    raise OptionalDependencyMissing('freediscovery_extra')
                cmod = make_xgboost_cv_model(cv, cv_obj, cv_scoring, **options)
        elif method == 'MLPClassifier':
            if cv is not None:
                raise NotImplementedFD('CV not supported with MLPClassifier')
            from sklearn.neural_network import MLPClassifier
            cmod = MLPClassifier(solver='adam', hidden_layer_sizes=10,
                                 max_iter=200, activation='identity',
                                 verbose=0,
                                 random_state=random_state)
        else:
            raise WrongParameter('Method {} not implemented!'.format(method))
        return cmod

    def fit(self, index, y, method='LinearSVC', cv=None):
        """
        Train the categorization model

        Parameters
        ----------
        index : array-like, shape (n_samples)
           document indices of the training set
        y : array-like, shape (n_samples)
           target class relative to index (string or int)
        method : str
           the ML algorithm to use (one of "LogisticRegression", "LinearSVC", 'xgboost')
        cv : str
           use cross-validation
        Returns
        -------
        cmod : sklearn.BaseEstimator
           the scikit learn classifier object
        Y_train : array-like, shape (n_samples)
           training predictions
        """

        valid_methods = ["LinearSVC", "LogisticRegression", "xgboost",
                         "NearestCentroid", "NearestNeighbor"]

        if method in ['MLPClassifier']:
            raise WrongParameter('method={} is implemented but not production ready. It was disabled for now.'.format(method))

        if method not in valid_methods:
            raise WrongParameter('method={} is not supported, should be one of {}'.format(
                method, valid_methods)) 
        if cv is not None and method in ['NearestNeighbor', 'NearestCentroid']:
            raise WrongParameter('Cross validation (cv={}) not supported with {}'.format(
                                        cv, method))

        if cv not in [None, 'fast', 'full']:
            raise WrongParameter('cv')

        d_all = self.pipeline.data

        X_train = d_all[index, :]

        Y_labels = y

        self.le = LabelEncoder()
        Y_train = self.le.fit_transform(Y_labels)

        cmod = self._build_estimator(Y_train, method, cv, self.cv_scoring, self.cv_n_folds)

        mid, mid_dir = setup_model(self.model_dir)

        if method == 'xgboost' and not cv:
            cmod.fit(X_train, Y_train, eval_metric='auc')
        else:
            cmod.fit(X_train, Y_train)

        joblib.dump(self.le, str(mid_dir / 'label_encoder'))
        joblib.dump(cmod, str(mid_dir / 'model'))

        pars = {
            'method': method,
            'index': index,
            'y': y,
            'categories': self.le.classes_
            }
        pars['options'] = cmod.get_params()
        self._pars = pars
        joblib.dump(pars, str(mid_dir / 'pars'))

        self.mid = mid
        self.cmod = cmod
        return cmod, Y_train

    def predict(self, chunk_size=5000, ml_output='probability', metric='cosine'):
        """
        Predict the relevance using a previously trained model

        Parameters
        ----------
        chunck_size : int
           chunk size
        ml_output : str
           type of the output in ['decision_function', 'probability'],
           only affects ML methods. default: 'probability'
        metric : str   
            The similarity returned by nearest neighbor classifier in
            ['cosine', 'jaccard', 'cosine_norm', 'jaccard_norm'].
            default: 'cosine'

        Returns
        -------
        res : ndarray [n_samples, n_classes]
           the score for each class
        nn_ind : {ndarray [n_samples, n_classes], None}
           the index of the nearest neighbor for each class
           (when the NearestNeighborRanker is used)
        """
        if ml_output not in ['probability', 'decision_function']:
            raise ValueError(("Wrong input value ml_output={}, must be one of "
                              "['probability', 'decision_function']")
                             .format(ml_output))

        if ml_output == 'probability':
            ml_output = 'predict_proba'

        if self.cmod is not None:
            cmod = self.cmod
        else:
            raise WrongParameter('The model must be trained first, or sid must be provided to load\
                    a previously trained model!')

        ds = self.pipeline.data

        nn_ind = None
        if isinstance(cmod, NearestNeighborRanker):
            res, nn_ind_orig = cmod.kneighbors(ds)
            res = _scale_cosine_similarity(res, metric=metric)
            nn_ind = self._pars['index'][nn_ind_orig]
        elif hasattr(cmod, ml_output):
            res = getattr(cmod, ml_output)(ds)
        elif hasattr(cmod, 'decision_function'):
            # and we need predict_proba
            res = cmod.decision_function(ds)
            res = expit(res)
        elif hasattr(cmod, 'predict_proba'):
            # and we need decision_function
            res = cmod.predict_proba(ds)
            res = logit(res)
        else:
            raise ValueError('Model {} has neither decision_function nor predict_proba methods!'.format(cmod))

        # handle the case of binary categorization
        # as two classes categorization
        if res.ndim == 1:
            if ml_output == 'decision_function':
                res_p = res
                res_n = - res
            else:
                res_p = res
                res_n = 1 - res
            res = np.hstack((res_n[:, None], res_p[:, None]))
        return res, nn_ind

    def _load_pars(self, mid=None):
        """Load model parameters from disk"""
        if mid is None:
            mid = self.mid
        mid_dir = self.model_dir / mid
        pars = super(_CategorizerWrapper, self)._load_pars(mid)
        cmod = joblib.load(str(mid_dir / 'model'))
        pars['options'] = cmod.get_params()
        return pars



# -*- coding: utf-8 -*-

from __future__ import print_function

from time import time

import numpy as np

from sklearn.preprocessing import normalize
from freediscovery.externals.birch import Birch
from freediscovery.cluster.birch import _BirchHierarchy, _print_container

from memory_profiler import memory_usage


n_samples = 100000
n_lsi = 150

np.random.seed(999999)

X = np.random.randn(n_samples, n_lsi)
normalize(X, norm='l2', copy=False)

print('Dataset with {} LSI dim'.format(n_lsi))

for N in np.linspace(0, n_samples, 5).astype(int)[1:]:


    t0 = time()
    mod = Birch(n_clusters=None, threshold=0.8, compute_labels=False)

    X_sl = X[:N, :]

    mod.fit(X_sl)
    t1 = time()
    bhmod = _BirchHierarchy(mod)
    bhmod.fit(X_sl)
    htree = bhmod.htree
    t2 = time()
    mem_usage = np.max(memory_usage())
    del mod
    del htree

    print('Data set {:9} docs, {:5.3f} GB RAM,  t_total= {:.2f} s, t_fit = {:.2f}s, t_hc = {:.2f}s'.format(
                        N, 1e-3*(mem_usage), t2-t0, t1-t0, t2 - t1))




import os
import sys
import time
import argparse
import shutil
from subprocess import call
from pathlib import Path

from sklearn.externals import joblib

from .server import fd_app
from freediscovery.engine.cli import (_query_yes_no, _TeeLogger,
                                      _number_of_workers)
from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.pipeline import PipelineFinder
from .datasets import IR_DATASETS, load_dataset
from ._version import __version__

DEFAULT_CACHE_DIR = os.path.join('..', 'freediscovery_shared')


def _parse_cache_dir(cache_dir):
    if cache_dir == DEFAULT_CACHE_DIR:
        cache_dir_new = os.environ.get('FREEDISCOVERY_CACHE_DIR')
        if cache_dir_new is not None:
            cache_dir = cache_dir
    return cache_dir


def _run(args):
    df_config = vars(args).copy()
    df_config.pop('func', None)
    df_config.pop('yes', None)

    cache_dir = _parse_cache_dir(args.cache_dir)
    cache_dir = os.path.normpath(os.path.abspath(cache_dir))
    if not os.path.exists(cache_dir):
        _cache_dir_exists = False
        _create_cache_dir = _query_yes_no('Cache directory does not exist. '
                                          'Create {} ?'.format(cache_dir),
                                          overwrite=args.yes)
        if _create_cache_dir:
            os.makedirs(cache_dir)
        else:
            print('Cache directory not created. Exiting.')
            return
    else:
        _cache_dir_exists = True

    df_config['cache_dir'] = cache_dir
    log_fname = args.log_file.replace('${CACHE_DIR}', cache_dir)
    log_fname = os.path.normpath(os.path.abspath(log_fname))

    df_config['log_file'] = log_fname
    df_config['n_workers'] = df_config.pop('n', None)

    # redirect stdout / stderr to a file
    sys.stdout = _TeeLogger(log_fname)
    print('='*80)
    print(' '*20, 'FreeDiscovery server')
    print(' '*21, '(version {})'.format(__version__))
    print('='*80)
    print(' * Started on {}'.format(time.strftime("%c")))
    print(' * CACHE_DIR: {} [{}]'.format(cache_dir,
                                         'EXISTING' if _cache_dir_exists else 'NEW'))
    print(' * LOG_FILE: {}'.format(log_fname))

    app = fd_app(cache_dir, df_config)
    if args.hostname == '0.0.0.0':
        print(' * WARNING: running the server on hostname 0.0.0.0 '
              '(accessible from any IP address). Please make sure '
              'that this server runs in a protected environement '
              '(e.g. is behind a firewall) or restrict '
              'connections to localhost with --hostname 127.0.0.1 .')

    if args.server in ['gunicorn']:
        try:
            from .server.gunicorn import GunicornApplication
            options = {
                'bind': '%s:%s' % (args.hostname, str(args.port)),
                'workers': args.n,
                'accesslog': '-',  # stdout
                'check_config': True,
                'limit_request_field_size': 0,  # unlimited
                'limit_request_line': 8190,
                'timeout': 18000,  # 5 hours
                'graceful_timeout': 18000,
            }
            parent_pid = os.getpid()
            print(' * Server: gunicorn with {} workers'.format(args.n))
            print(' * Running on http://{}/ (Press CTRL+C to quit)'.format(options['bind']))
            GunicornApplication(app, options).run()
            return
        except SystemExit:
            if os.getpid() == parent_pid:
                print("Stopping FreeDiscovery server.")
            return
        except ImportError:
            #print('Gunicorn not installed')
            pass
        except:
            if os.getpid() == parent_pid:
                print('Exiting.')
            return

    # run the built-in server
    print(' * Server: flask (Werkzeug) threaded')

    df_config['server'] = 'werkzeug'
    app.run(debug=False, host=args.hostname,
            processes=1, threaded=True,
            port=args.port, use_reloader=False)


def _info(args):

    print('FreeDiscovery version: {}'.format(__version__))

    conda_exec = shutil.which('conda')
    pip_exec = shutil.which('pip')

    if conda_exec is not None:
        print("\n# Conda setup\n")
        if args.all:
            call([conda_exec, 'info', '-a'])
            print('\n# List of installed packages\n')
            call([conda_exec, 'list'])
        else:
            call([conda_exec, 'info'])
    elif pip_exec is not None:
        print("\n# Python setup\n")
        call([pip_exec, '-V'])
        print('\n# List of installed packages\n')
        call([pip_exec, 'list'])
    else:
        print('Warning: neither pip nor conda found in $PATH!')


def _list(args):
    cache_dir = _parse_cache_dir(args.cache_dir)
    fe = FeatureVectorizer(cache_dir)
    res = fe.list_datasets()
    res = sorted(res, key=lambda row: row['creation_date'], reverse=True)
    for row in res:
        print(' * Processed dataset {}'.format(row['id']))
        print('    - data_dir: {}'.format(row['data_dir']))
        print('    - creation_date: {}'.format(row['creation_date']))
        for method in ['lsi', 'categorizer', 'cluster', 'dupdet',
                       'threading']:
            dpath = os.path.join(fe.cache_dir, row['id'], method)
            if not os.path.exists(dpath):
                continue
            mid_list = os.listdir(dpath)
            if mid_list:
                print('     # {}'.format(method))
                for mid in mid_list:
                    print('       * {}'.format(mid))
        print(' ')


def _show(args):
    cache_dir = _parse_cache_dir(args.cache_dir)
    p = PipelineFinder.by_id(mid=args.mid, cache_dir=cache_dir)
    print(p)
    print(' * model_id: {}'.format(args.mid))
    print(' * model_type: {}'.format(list(p.keys())[-1]))
    print(' * file_path: {}'.format(p.get_path()))
    try:
        pars = joblib.load(os.path.join(p.get_path(), 'pars'))
        for key, val in pars.items():
            val_str = str(val)
            if len(val_str) > 30 and not isinstance(val, dict):
                continue
            print(' * {}: {}'.format(key, val_str))
    except:
        pass


def _rm(args):
    cache_dir = _parse_cache_dir(args.cache_dir)
    if args.all:
        p = PipelineFinder(cache_dir=cache_dir)
        fpath = p.cache_dir
    elif args.mid:
        mid = args.mid
        p = PipelineFinder.by_id(mid=mid, cache_dir=cache_dir)
        fpath = p.get_path()
    else:
        print('Error: either mid or the -a (--all) flag should be provided. '
              'Exiting.')
        return
    _del_mid = _query_yes_no('Are you sure you want to delete\n'
                             '        {} ?'.format(fpath),
                             default='no',
                             overwrite=args.yes)
    if _del_mid:
        shutil.rmtree(fpath)
        print('Folder {} deleted.'.format(fpath))
    else:
        print('Nothing to be done. Exiting.')

def _download(args):
    cache_dir = Path(args.output).resolve()

    if not cache_dir.exists():
        raise ValueError('Output directory {} does not exist!'
                         .format(cache_dir))

    load_dataset(args.name, cache_dir=cache_dir, verbose=True)


# Allow to propagate formatter_class to subparsers
# https://bugs.python.org/issue21633

class _ArgParser(argparse.ArgumentParser):

    def __init__(self, **kwargs):
        kwargs["formatter_class"] = argparse.ArgumentDefaultsHelpFormatter
        super(_ArgParser, self).__init__(**kwargs)


def main(args=None, return_parser=False):
    """The main CLI interface."""

    parser = _ArgParser()
    subparsers = parser.add_subparsers(help='action')
    # parser.add_argument("-v", ..)

    run_parser = subparsers.add_parser("run",
                     description='The command used to start the server.')
    info_parser = subparsers.add_parser("info",
                     description='Return debug information about '
                                 'the FreeDiscovery install.')
    list_parser = subparsers.add_parser("list",
                     description='List trained models.')
    show_parser = subparsers.add_parser("show", 
                     description='Show detailed information about '
                                 'a trained model.')
    rm_parser = subparsers.add_parser("rm", 
                     description='Remove a trained model specified by its ID.')
    download_parser = subparsers.add_parser("download", 
                     description='Download a document dataset')

    for subparser in [run_parser, list_parser, show_parser,
                      rm_parser]:
        subparser.add_argument('-c', '--cache-dir',
                               default=DEFAULT_CACHE_DIR,
                               help='The cache directory in which '
                                    'the trained models are saved. '
                                    'If this parameter is not specified '
                                    'the value in the environement variable '
                                    'FREEDISCOVERY_CACHE_DIR will be used if '
                                    'it is specified. Otherwise the default '
                                    'value is used.')
    for subparser in [run_parser, rm_parser]:
        subparser.add_argument('-y', '--yes', action='store_true',
                               help='Do not ask for confirmation.')

    # start parser
    run_parser.add_argument('--debug',
                            default=False, action='store_true',
                            help='Start server in debug mode.')
    run_parser.add_argument('--hostname', default='0.0.0.0',
                            help='Server hostname.')
    run_parser.add_argument('-p', '--port', default=5001, type=int,
                            help='Server port.')
    run_parser.add_argument('-s', '--server', default='flask',
                            choices=['flask', 'gunicorn'],
                            help='The server used to run freediscovery. '
                                 '"flask" is the server built-in in flask '
                                 'suitable for developpement. ')
    run_parser.add_argument('--log-file',
                            default='${CACHE_DIR}/freediscovery-backend.log',
                            help='Path to the log file.')
    run_parser.add_argument('-n', default=_number_of_workers(), type=int,
                            help='Number of workers to use when starting '
                                 'the freediscovery server. Only affects '
                                 'the gunicorn server.')
    run_parser.set_defaults(func=_run)

    # info parser
    info_parser.add_argument('-a', '--all', action='store_true',
                             help='Print detailed report.')
    info_parser.set_defaults(func=_info)

    # list parser
    list_parser.set_defaults(func=_list)

    # show parser
    show_parser.add_argument('mid', help='Model id')
    show_parser.set_defaults(func=_show)

    # rm parser
    rm_parser.add_argument('-a', '--all', action='store_true',
                           help='Remove all models.')
    rm_parser.add_argument('mid', nargs='?', help='Model id')
    rm_parser.set_defaults(func=_rm)

    # download parser
    download_parser.add_argument('-o', '--output',
                             default='.',
                             help='Folder where to save the output folder')
    download_parser.add_argument('name',
                             help='The dataset name',
                             choices=IR_DATASETS.keys())
    download_parser.set_defaults(func=_download)

    if return_parser:
        # used to generate sphinx docs
        return parser

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os.path
import shutil
import pickle
import warnings
import time

import numpy as np
import pandas as pd
import scipy.sparse
from sklearn.externals import joblib
from sklearn.externals.joblib import Parallel, delayed
from sklearn.feature_extraction.text import CountVectorizer

from freediscovery._version import __version__
from freediscovery.engine.pipeline import PipelineFinder
from freediscovery.engine.stop_words import _StopWordsWrapper
from freediscovery.engine.ingestion import DocumentIndex
from freediscovery.utils import generate_uuid, _rename_main_thread
from freediscovery.feature_weighting import SmartTfidfTransformer, _validate_smart_notation
from freediscovery.preprocessing import processing_filters
from freediscovery.exceptions import (DatasetNotFound, InitException, WrongParameter)
from freediscovery.engine.utils import validate_mid


def _touch(filename):
    filename.open('ab').close()


def _read_file(file_path):
    with open(file_path, 'rb') as fh:
        doc = fh.read()
    doc = doc.decode('utf-8', 'ignore')
    return doc


def _preprocess_stream(doc, steps=None):
    """ Apply pre-processing steps """

    if steps:
        for key in steps:
            func = processing_filters[key]
            doc = func(doc)

    return doc


def _vectorize_chunk(dsid_dir, k, pars, pretend=False):
    """ Extract features on a chunk of files """
    from sklearn.feature_extraction.text import HashingVectorizer
    from sklearn.externals import joblib

    filenames = pars['filenames_abs']
    chunk_size = pars['chunk_size']
    n_samples = pars['n_samples']

    mslice = slice(k*chunk_size, min((k+1)*chunk_size, n_samples))

    hash_opts = {key: vals for key, vals in pars.items()
                 if key in ['stop_words', 'n_features',
                            'analyser', 'ngram_range']}
    hash_opts['alternate_sign'] = False
    fe = HashingVectorizer(input='content', norm=None, **hash_opts)
    if pretend:
        return fe
    fset_new = fe.transform(_read_file(fname) for fname in filenames[mslice])

    fset_new.eliminate_zeros()

    joblib.dump(fset_new, str(dsid_dir / 'features-{:05}'.format(k)))


class FeatureVectorizer(object):
    """Extract features from text documents

    Parameters
    ----------
    cache_dir : str, default='/tmp/'
        directory where to save temporary and regression files
    dsid : str
        load an exising dataset
    verbose : bool
        pring progress messages
    mode : str
        write or read mode for the FeatureVectorizer
    """

    _PARS_SHORT = ['data_dir', 'n_samples', 'n_features',
                   'n_jobs', 'chunk_size',
                   'analyzer', 'ngram_range', 'stop_words',
                   'weighting', 'norm_alpha', 'use_hashing',
                   'creation_date']

    _wrapper_type = "vectorizer"

    def __init__(self, cache_dir='/tmp/', dsid=None, verbose=False, mode='r'):
        self.verbose = verbose

        self._filenames = None
        self._vect = None
        self._tfidf = None
        self._db = None
        self._pars = None

        self.cache_dir = cache_dir = PipelineFinder._normalize_cachedir(cache_dir)
        if not cache_dir.exists():
            cache_dir.mkdir()
        self.dsid = dsid
        if mode not in ['r', 'w', 'fw']:
            raise WrongParameter('mode={} must be one of "r", "w", "fw"'
                                 .format(mode))
        self.mode = mode
        if dsid is not None:
            validate_mid(dsid)
            dsid_dir = self.cache_dir / dsid
            if mode == 'r':
                if not dsid_dir.exists():
                    raise DatasetNotFound('Dataset '
                                          '{} ({}) not found in {}!'.format(
                                           dsid, type(self).__name__, cache_dir))
            else:
                if dsid_dir.exists():
                    if mode == 'w':
                        raise WrongParameter(('dataset identified by dsid={} '
                                              'already exists. Use mode="fw" '
                                              'to overwrite.')
                                             .format(dsid))
                    elif mode == 'fw':
                        shutil.rmtree(dsid_dir)
        else:
            dsid_dir = None
        self.dsid_dir = dsid_dir

    @property
    def n_samples_(self):
        """ Number of documents in the dataset """
        return self.pars_['n_samples']

    @property
    def filenames_(self):
        """ Lazily load the list of filenames if needed """
        if not hasattr(self, '_filenames') or self._filenames is None:
            with (self.dsid_dir / 'filenames').open('rb') as fh:
                self._filenames = pickle.load(fh)
        return self._filenames

    @property
    def pars_(self):
        if not hasattr(self, '_pars') or self._pars is None:
            # Load parameters from disk
            dsid = self.dsid
            if self.cache_dir is None:
                raise InitException('cache_dir is None: '
                                    'cannot load from cache!')
            dsid_dir = self.cache_dir / dsid
            if not dsid_dir.exists():
                raise DatasetNotFound('dsid {} not found!'.format(dsid))
            with (dsid_dir / 'pars').open('rb') as fh:
                self._pars = pickle.load(fh)
        return self._pars

    @property
    def vect_(self):
        if not hasattr(self, '_vect') or self._vect is None:
            mid = self.dsid
            mid_dir = self.cache_dir / mid
            if not mid_dir.exists():
                raise ValueError(('Vectorizer model id {} ({}) '
                                  'not found in the cache {}!')
                                 .format(mid, mid_dir))
            fname = mid_dir / 'vectorizer'
            if self.pars_['use_hashing']:
                self._vect = joblib.load(str(fname))
            else:
                # this is much faster in python 3 as cpickle is used
                # (only works if no numpy arrays are used)
                with fname.open('rb') as fh:
                    self._vect = pickle.load(fh)
        return self._vect

    @property
    def tfidf_(self):
        if not hasattr(self, '_tfidf') or self._tfidf is None:
            mid = self.dsid
            mid_dir = self.cache_dir / mid
            if not mid_dir.exists():
                raise ValueError(('Vectorizer model id {} ({}) '
                                  'not found in the cache {}!')
                                 .format(mid, mid_dir))
            fname = mid_dir / 'tfidf_transformer'
            self._tfidf = joblib.load(str(fname))
        return self._tfidf

    def _load_features(self, dsid=None):
        """ Load a computed features from disk"""
        if self.cache_dir is None:
            raise InitException('cache_dir is None: cannot load from cache!')
        if dsid is None:
            dsid = self.dsid
        dsid_dir = self.cache_dir / dsid
        if not dsid_dir.exists():
            raise DatasetNotFound('dsid not found!')
        fset_new = joblib.load(str(dsid_dir / 'features'))
        return fset_new

    def setup(self, n_features=None, chunk_size=5000, analyzer='word',
              ngram_range=(1, 1), stop_words=None, n_jobs=1,
              use_hashing=False,
              weighting='nnc', norm_alpha=0.75, min_df=0.0, max_df=1.0,
              parse_email_headers=False,
              preprocess=[]):
        """Initalize the features extraction.

        See sklearn.feature_extraction.text for a detailed description
        of the input parameters

        Parameters
        ----------
        analyzer : string, {'word', 'char'} or callable
            Whether the feature should be made of word or character n-grams.
            If a callable is passed it is used to extract the sequence of
            features out of the raw, unprocessed input.
        ngram_range : tuple (min_n, max_n)
            The lower and upper boundary of the range of n-values for different
            n-grams to be extracted. All values of n such that
            min_n <= n <= max_n will be used.
        stop_words : string {'english'}, list, or None (default)
            If a string, it is passed to _check_stop_list and the appropriate
            stop list is returned. 'english' is currently the only supported
            string value.
            If None, no stop words will be used. max_df can be set to a value
            in the range [0.7, 1.0) to automatically detect and filter stop
            words based on intra corpus document frequency of terms.
        max_df : float in range [0.0, 1.0] or int, default=1.0
            When building the vocabulary ignore terms that have a document
            frequency strictly higher than the given threshold (corpus-specific
            stop words).
            If float, the parameter represents a proportion of documents,
            integer absolute counts.
            This parameter is ignored if vocabulary is not None.
        min_df : float in range [0.0, 1.0] or int, default=1
            When building the vocabulary ignore terms that have a document
            frequency strictly lower than the given threshold. This value is
            also called cut-off in the literature.
            If float, the parameter represents a proportion of documents,
            integer absolute counts.
            This parameter is ignored if vocabulary is not None.
        max_features : int or None, default=None or 100001
            If not None, build a vocabulary that only consider the top
            max_features ordered by term frequency across the corpus.
        weighting : str
            SMART weighting type
        preprocess : list of strings, default: []
            A list of pre-processing steps, including 'emails_ingore_header'
        """
        if self.mode not in ['w', 'fw']:
            raise WrongParameter('The vectorizer can be setup only with '
                                 'mode in ["w", "fw"]')

        if analyzer not in ['word', 'char', 'char_wb']:
            raise WrongParameter('analyzer={} not supported!'.format(analyzer))

        if not isinstance(ngram_range, tuple) \
           and not isinstance(ngram_range, list):
            raise WrongParameter(('not a valid input ngram_range='
                                  '{}: should be a list or a typle!')
                                 .format(ngram_range))

        if not len(ngram_range) == 2:
            raise WrongParameter('len(gram_range=={}!=2'
                                 .format(len(ngram_range)))

        if not 0 <= norm_alpha <= 1:
            raise WrongParameter('norm_alpha={} not in [0, 1]'
                                 .format(norm_alpha))

        _, _, weighting_n = _validate_smart_notation(weighting)
        if weighting_n == 'n':
            warnings.warn('You should use either cosine or pivoted normalization '
                          'i.e. weighting should be "**[cp]"',
                          UserWarning)

        for key in preprocess:
            if key not in processing_filters:
                raise WrongParameter(('Unknown preprocessing step {} '
                                      ' must of be of {}')
                                     .format(key, ', '.join(list(processing_filters.keys()))))

        if stop_words in [None, 'english', 'english_alphanumeric']:
            pass
        elif stop_words in _StopWordsWrapper(cache_dir=self.cache_dir):
            pass
        else:
            raise WrongParameter('stop_words = {}'.format(stop_words))

        if n_features is None and use_hashing:
            n_features = 100001  # default size of the hashing table

        if self.dsid is None:
            self.dsid = dsid = generate_uuid()
        else:
            dsid = self.dsid
        self.dsid_dir = dsid_dir = self.cache_dir / dsid

        dsid_dir.mkdir()

        pars = {'data_dir': None,
                'n_samples': None, "n_features": n_features,
                'chunk_size': chunk_size, 'stop_words': stop_words,
                'analyzer': analyzer, 'ngram_range': ngram_range,
                'n_jobs': n_jobs, 'use_hashing': use_hashing,
                'weighting': weighting, 'norm_alpha': norm_alpha,
                'min_df': min_df, 'max_df': max_df,
                'parse_email_headers': parse_email_headers,
                'type': type(self).__name__,
                'preprocess': preprocess,
                'freediscovery_version': __version__}
        self._pars = pars
        with (dsid_dir / 'pars').open('wb') as fh:
            pickle.dump(self._pars, fh)
        return dsid

    def ingest(self, data_dir=None, file_pattern='.*', dir_pattern='.*',
               dataset_definition=None, vectorize=True,
               document_id_generator='indexed_file_path'):
        """Perform data ingestion

        Parameters
        ----------
        data_dir : str
            path to the data directory (used only if metadata not provided),
            default: None
        dataset_defintion : list of dicts
            a list of dictionaries with keys
            ['file_path', 'document_id', 'rendition_id']
            describing the data ingestion (this overwrites data_dir)
        vectorize : bool (default: True)
        """
        dsid_dir = self.cache_dir / self.dsid
        if (dsid_dir / 'db').exists():
            raise ValueError('Dataset {} already vectorized!'
                             .format(self.dsid))
        db_list = list(sorted(dsid_dir.glob('db*')))
        if len(db_list) == 0:
            internal_id_offset = -1
        elif len(db_list) >= 1:
            internal_id_offset = int(db_list[-1].name[3:])

        if dataset_definition is not None:
            db = DocumentIndex.from_list(dataset_definition, data_dir,
                                         internal_id_offset + 1, dsid_dir,
                                         document_id_generator=document_id_generator)
        elif data_dir is not None:
            db = DocumentIndex.from_folder(data_dir, file_pattern, dir_pattern,
                                           internal_id_offset + 1,
                                           document_id_generator=document_id_generator)
        else:
            db = None

        if db is not None:
            data_dir = db.data_dir

            batch_suffix = '.{:09}'.format(db.data.internal_id.iloc[-1])

            self._filenames = db.data.file_path.values.tolist()
            del db.data['file_path']

            pars = self.pars_

            if 'file_path' in db.data.columns:
                del db.data['file_path']
            db.data.to_pickle(str(dsid_dir / ('db' + batch_suffix)))
            with (dsid_dir / ('filenames' + batch_suffix)).open('wb') as fh:
                pickle.dump(self._filenames, fh)
            self._db = db

        if vectorize:
            db_list = list(sorted(dsid_dir.glob('db*')))
            filenames_list = list(sorted(dsid_dir.glob('filenames*')))
            if len(db_list) == 0:
                raise ValueError('No ingested files found!')

            if len(db_list) == 1:
                with filenames_list[0].open('rb') as fh:
                    filenames_concat = pickle.load(fh)
            elif len(db_list) >= 2:
                # accumulate different batches into a single file
                # filename file
                filenames_concat = []
                for fname in filenames_list:
                    with fname.open('rb') as fh:
                        filenames_concat += pickle.load(fh)

            if self.pars_['data_dir'] is None:
                data_dir = DocumentIndex._detect_data_dir(filenames_concat)
                self._pars['data_dir'] = data_dir
            else:
                data_dir = self._pars['data_dir']

            self._filenames = [os.path.relpath(el, data_dir)
                               for el in filenames_concat]

            with (dsid_dir / 'filenames').open('wb') as fh:
                pickle.dump(self._filenames, fh)

            for fname in filenames_list:
                fname.unlink()

            # save databases
            if len(db_list) == 1:
                db_list[0].rename(dsid_dir / 'db')
                self.db_.filenames_ = self._filenames
                self.db_.data['file_path'] = self._filenames
            elif len(db_list) >= 2:

                db_concat = []
                for fname in db_list:
                    db_concat.append(pd.read_pickle(str(fname)))
                db_new = pd.concat(db_concat, axis=0)
                db_new.filenames_ = self._filenames
                db_new.set_index('internal_id', drop=False, inplace=True)
                self._db = DocumentIndex(data_dir, db_new)
                if 'file_path' in db_new.columns:
                    del db_new['file_path']
                db_new.to_pickle(str(dsid_dir / 'db'))

            # save parameters
            self._pars['n_samples'] = len(self._filenames)
            self._pars['data_dir'] = data_dir

            with (dsid_dir / 'pars').open('wb') as fh:
                pickle.dump(self._pars, fh)

            self.transform()

            if (dsid_dir / 'raw').exists():
                shutil.rmtree(str(dsid_dir / 'raw'))

        if db is None and not vectorize:
            raise ValueError('At least one of data_dir, dataset_definition, '
                             'vectorize parameters must be provided!')
        return

    @staticmethod
    def _generate_stop_words(stop_words):
        if stop_words in [None]:
            return None
        elif stop_words == 'english':
            return stop_words
        else:
            raise ValueError

    def parse_email_headers(self):
        from email.parser import Parser
        from freediscovery.externals.jwzthreading import Message
        features = []
        for idx, fname in enumerate(self.filenames_abs_):
            txt = _read_file(fname)
            message = Parser().parsestr(txt, headersonly=True)

            msg_obj = Message(message, message_idx=idx)

            features.append(msg_obj)
        joblib.dump(features, str(self.dsid_dir / 'email_metadata'))
        return features

    @property
    def filenames_abs_(self):
        """ Return the absolute path to filenames """
        data_dir = self.pars_['data_dir']
        return [os.path.join(data_dir, el) for el in self.filenames_]

    def transform(self):
        """
        Run the feature extraction
        """
        dsid_dir = self.dsid_dir
        if not dsid_dir.exists():
            raise DatasetNotFound()

        if not (dsid_dir / 'db').exists():
            raise ValueError('Please ingest some files before running '
                             'the vectorizer!')

        pars = self.pars_
        pars['filenames_abs'] = self.filenames_abs_
        chunk_size = pars['chunk_size']
        n_samples = pars['n_samples']
        use_hashing = pars['use_hashing']

        if use_hashing:
            # make sure that we can initialize the vectorizer
            # (easier outside of the paralel loop
            vect = _vectorize_chunk(dsid_dir, 0, pars, pretend=True)

        processing_lock = (dsid_dir / 'processing')
        _touch(processing_lock)
        custom_sw = _StopWordsWrapper(cache_dir=self.cache_dir)
        if pars['stop_words'] in custom_sw:
            pars['stop_words'] = custom_sw.load(pars['stop_words'])
        else:
            pars['stop_words'] = self._generate_stop_words(pars['stop_words'])

        try:
            if use_hashing:
                # fixed in https://github.com/joblib/joblib/pull/414
                _rename_main_thread()
                Parallel(n_jobs=pars['n_jobs'])(
                            delayed(_vectorize_chunk)(dsid_dir, k, pars)
                            for k in range(n_samples//chunk_size + 1))

                res = self._aggregate_features()

                self._vect = vect
            else:
                opts_tfidf = {key: val for key, val in pars.items()
                              if key in ['stop_words',
                                         'ngram_range', 'analyzer',
                                         'min_df', 'max_df']}

                vect = CountVectorizer(input='content',
                                       max_features=pars['n_features'],
                                       **opts_tfidf)
                text_gen = (_preprocess_stream(_read_file(fname), pars['preprocess'])
                            for fname in pars['filenames_abs'])
                res = vect.fit_transform(text_gen)
                self._vect = vect
            fname = dsid_dir / 'vectorizer'
            if self._pars['use_hashing']:
                joblib.dump(self._vect, str(fname))
            else:
                # faster for pure python objects
                with fname.open('wb') as fh:
                    pickle.dump(self._vect, fh)
            fname = dsid_dir / 'tfidf_transformer'
            wt = SmartTfidfTransformer(pars['weighting'],
                                       norm_alpha=pars['norm_alpha'])
            self._idf = wt
            res = wt.fit_transform(res)
            joblib.dump(self._idf, str(fname))

            del self.pars_['filenames_abs']

            joblib.dump(res, str(dsid_dir / 'features'))
            # remove all identical files
            if use_hashing:
                for filename in dsid_dir.glob('features-*[0-9]*'):
                    filename.unlink()
        except:
            if processing_lock.exists():
                processing_lock.unlink()
            raise
        # remove processing lock if finished or if error
        if processing_lock.exists():
            processing_lock.unlink()
        _touch(dsid_dir / 'processing_finished')

    def append(self, dataset_definition, data_dir=None):
        """ Add some documents to the dataset

        This is by no mean an efficient operation, processing all the files
        at once might be more suitable in most occastions.
        """
        from freediscovery.engine.lsi import _LSIWrapper
        dsid_dir = self.dsid_dir
        db_old = self.db_.data
        internal_id_offset = db_old.internal_id.max()
        db_extra = DocumentIndex.from_list(dataset_definition, data_dir,
                                           internal_id_offset + 1, dsid_dir)
        db_new = db_extra.data
        vect = self.vect_
        tfidf = self.tfidf_

        filenames_new = list(db_new.file_path.values)

        # write down the new features file
        X_new_raw = vect.transform(filenames_new)
        X_new = tfidf.transform(X_new_raw)
        X_old = self._load_features()
        X = scipy.sparse.vstack((X_new, X_old))
        joblib.dump(X, str(dsid_dir / 'features'))

        # write down the new filenames file
        filenames_old = list(self.filenames_)
        filenames = filenames_old + filenames_new

        data_dir = DocumentIndex._detect_data_dir(filenames)
        self._pars['data_dir'] = data_dir

        self._filenames = [os.path.relpath(el, data_dir)
                           for el in filenames]

        with (dsid_dir / 'filenames').open('wb') as fh:
            pickle.dump(self._filenames, fh)
        del db_new['file_path']

        # write down the new pars file
        self._pars = self.pars_
        self._pars['n_samples'] = len(filenames)
        with (dsid_dir / 'pars').open('wb') as fh:
            pickle.dump(self._pars, fh)

        # write down the new database file
        db = pd.concat((db_old, db_new))
        if 'file_path' in db.columns:
            del db['file_path']
        db.to_pickle(str(dsid_dir / 'db'))
        self._db = DocumentIndex(self.pars_['data_dir'], db)

        # find all exisisting LSI models and update them as well
        if (dsid_dir / 'lsi').exists():
            for lsi_id in os.listdir(str(dsid_dir / 'lsi')):
                lsi_obj = _LSIWrapper(cache_dir=self.cache_dir,
                                      mid=lsi_id)
                lsi_obj.append(X_new)

        # remove all trained models for this dataset
        for model_type in ['categorizer', 'dupdet', 'cluster', 'threading']:
            if (dsid_dir / model_type).exists():
                for mid in os.listdir(str(dsid_dir / model_type)):
                    shutil.rmtree(str(dsid_dir / model_type / mid))

    def remove(self, dataset_definition):
        """ Remove some documents from the dataset

        This is by no mean an efficient operation, processing all the files
        at once might be more suitable in most occastions.
        """
        from freediscovery.engine.lsi import _LSIWrapper
        dsid_dir = self.dsid_dir
        db_old = self.db_.data
        query = pd.DataFrame(dataset_definition)
        res = self.db_.search(query, drop=False)
        del_internal_id = res.internal_id.values
        internal_id_mask = ~np.in1d(db_old.internal_id.values, del_internal_id)

        # write down the new features file
        X_old = self._load_features()
        X = X_old[internal_id_mask, :]
        joblib.dump(X, str(dsid_dir / 'features'))

        # write down the new filenames file
        filenames = list(np.array(self.filenames_)[internal_id_mask])
        with (dsid_dir / 'filenames').open('wb') as fh:
            pickle.dump(filenames, fh)
        self._filenames = filenames

        # write down the new database file
        db = db_old.iloc[internal_id_mask].copy()
        # create a new contiguous internal_id
        db['internal_id'] = np.arange(db.shape[0], dtype='int')
        self._db = DocumentIndex(self.pars_['data_dir'], db)
        if 'file_path' in db.columns:
            del db['file_path']
        db.to_pickle(str(dsid_dir / 'db'))

        # write down the new pars file
        self._pars = self.pars_
        self._pars['n_samples'] = len(filenames)
        with (dsid_dir / 'pars').open('wb') as fh:
            pickle.dump(self._pars, fh)

        # find all exisisting LSI models and update them as well
        if (dsid_dir / 'lsi').exists():
            for lsi_id in os.listdir(str(dsid_dir / 'lsi')):
                _fname = dsid_dir / 'lsi' / lsi_id / 'data'
                if _fname.exists():
                    X_lsi_old = joblib.load(str(_fname))
                    X_lsi = X_lsi_old[internal_id_mask]
                    joblib.dump(X_lsi, str(_fname))

        # remove all trained models for this dataset
        for model_type in ['categorizer', 'dupdet', 'cluster', 'threading']:
            if (dsid_dir / model_type).exists():
                for mid in os.listdir(str(dsid_dir / model_type)):
                    shutil.rmtree(str(dsid_dir / model_type / mid))

    @property
    def n_features_(self):
        """ Number of features of the vecotorizer"""
        from sklearn.feature_extraction.text import HashingVectorizer
        vect = self.vect_
        if hasattr(vect, 'vocabulary_'):
            return len(vect.vocabulary_)
        elif isinstance(vect, HashingVectorizer):
            return vect.get_params()['n_features']
        else:
            raise ValueError

    def query_features(self, indices, n_top_words=10, remove_stop_words=False):
        """ Query the features with most weight

        Parameters
        ----------
        indices : list or ndarray
          indices for the subcluster
        n_top_words : int
          the number of workds to return
        remove_stop_words : bool
          remove stop words
        """
        from freediscovery.utils import _query_features

        X = self._load_features()
        return _query_features(self.vect_, X, indices, n_top_words,
                               remove_stop_words)

    def _aggregate_features(self):
        """ Agregate features loaded as separate files features-<number>
        into a single file features"""
        from glob import glob
        out = []
        for filename in sorted(self.dsid_dir.glob('features-*[0-9]')):
            ds = joblib.load(str(filename))
            out.append(ds)
        res = scipy.sparse.vstack(out)
        return res

    @property
    def db_(self):
        """ DatasetIndex """
        if not hasattr(self, '_db') or self._db is None:
            dsid = self.dsid
            if self.cache_dir is None:
                raise InitException('cache_dir is None: cannot load from cache!')
            dsid_dir = self.cache_dir / dsid
            if not dsid_dir.exists():
                raise DatasetNotFound('dsid {} not found!'.format(dsid))
            data = pd.read_pickle(str(dsid_dir / 'db'))
            self._db = DocumentIndex(self.pars_['data_dir'], data)
        return self._db

    def delete(self):
        """ Delete the current dataset """
        import shutil
        shutil.rmtree(self.dsid_dir, ignore_errors=True)

    def __contains__(self, dsid):
        """ This is a somewhat non standard call that checks if a dsid
        exist on disk (in general)"""
        return (self.cache_dir / dsid).exists()

    def __getitem__(self, index):
        return np.asarray(self.filenames_)[index]

    def list_datasets(self):
        """ List all datasets in the working directory """
        import traceback
        out = []
        for dsid in os.listdir(str(self.cache_dir)):
            if dsid == 'stop_words':
                continue
            row = {"id": dsid}
            self.dsid = dsid
            try:
                pars = self.pars_
            except:
                # traceback.print_exc()
                continue

            if pars['type'] != type(self).__name__:
                continue

            creation_date = os.stat(str(self.cache_dir / dsid)).st_ctime
            pars['creation_date'] = time.strftime('%c',
                                                  time.gmtime(creation_date))

            try:
                for key in self._PARS_SHORT:
                    row[key] = pars[key]
                out.append(row)
            except Exception:
                print(pars.keys())
                traceback.print_exc()

        return out


class _FeatureVectorizerSampled(FeatureVectorizer):
    """Extract features from text documents, with additional sampling

    This is mostly a helper class for debugging and experimenting

    Parameters
    ----------
    cache_dir : str, default='/tmp/'
        directory where to save temporary and regression files
    dsid : str
        load an exising dataset
    verbose : bool
        pring progress messages
    sampling_index : list of filenames or None, default=None
        a list of filenames in self._pars['filenames']
        (with possible duplicates) used to downsample / upsample the dataset

    """
    def __init__(self, cache_dir='/tmp/', dsid=None, verbose=False,
                 sampling_filenames=None):
        super(_FeatureVectorizerSampled, self).__init__(cache_dir, dsid,
                                                        verbose)

        if sampling_filenames is not None:
            self.sampling_filenames = sampling_filenames
            if dsid is None:
                raise ValueError('sampling_index must be applied '
                                 'onto an existing dataset\n'
                                 'specified by dsid')
            if not isinstance(sampling_filenames, list):
                raise TypeError('Wrong type {} for sampling_index,'
                                ' must be list'.format(
                                    type(sampling_filenames).__name__))

            self.db_.filenames_ = super(_FeatureVectorizerSampled, self).filenames_
            self.sampling_index = self.db_._search_filenames(self.sampling_filenames)
        else:
            self.sampling_filenames = None
            self.sampling_index = None

    @property
    def pars_(self):
        pars = super(_FeatureVectorizerSampled, self).pars_
        if self.sampling_filenames is not None:
            pars['n_samples'] = self.n_samples_
            pars['additional_sampling'] = True
        else:
            pars['additional_sampling'] = False
        return pars

    def _load_features(self, dsid):
        X = super(_FeatureVectorizerSampled, self)._load_features(dsid)
        if self.sampling_filenames is not None:
            return X[self.sampling_index, :]
        else:
            return X

    @property
    def filenames_(self):
        fnames = super(_FeatureVectorizerSampled, self).filenames_
        fnames = np.array(fnames)
        if self.sampling_filenames is not None:
            fnames = fnames[self.sampling_index]
        return fnames

    @property
    def n_samples_(self):
        if self.sampling_filenames is not None:
            return len(self.sampling_filenames)
        else:
            return super(_FeatureVectorizerSampled, self).n_samples_



import pandas as pd
import numpy as np
from time import time

from freediscovery.ingestion import DocumentIndex


def test_performance():

    md = []
    t0 = time()

    for idx in range(10000):
        md.append({'file_path': '/test{}'.format(idx),
                   'internal_id': idx,
                   'document_id': idx**2})
    query = []
    for idx in range(10000):
        query.append({'internal_id': idx, 
                   'a': np.random.randint(100)
                   })
    query = pd.DataFrame(query)

    dbi = DocumentIndex.from_list(md)
    print(time() - t0)

    t0 = time()
    rd = dbi.render_dict(query)
    print(time() - t0)
    print(len(rd))


test_performance()



# -*- coding: utf-8 -*-

from numpy.testing import assert_array_less

import pytest

from sklearn.feature_extraction.text import CountVectorizer
from freediscovery.feature_weighting import SmartTfidfTransformer

from freediscovery.lsi import _TruncatedSVD_LSI
from freediscovery.search import Search


@pytest.mark.parametrize('kind,', ['regular', 'semantic'])
def test_search(kind):
    # testing that search algorithm actually works
    corpus = ["To be, or not to be; that is the question;",
              "Whether ‘tis nobler in the mind to suffer",
              "The slings and arrows of outrageous fortune,",
              "Or to take arms against a sea of troubles,",
              "And by opposing end them. To die: to sleep:",
              "Nor more; and by a sleep to say we end",
              "The heart-ache and the thousand natural shocks",
              "That flesh is heir to; ‘tis a consummation",
              "Devoutly to be wished. To die; to sleep;",
              "To sleep: perchance to dream: aye, there is the rub;",
              "For in that sleep of death what dreams may come,",
              "When we have shuffled off this mortal coil,",
              "Must give us pause: there’s the respect",
              "That makes calamity of so long life;"]

    vect = CountVectorizer()
    X_tf = vect.fit_transform(corpus)
    idf = SmartTfidfTransformer('nnc')
    X_vect = idf.fit_transform(X_tf)

    if kind == 'semantic':
        lsi = _TruncatedSVD_LSI(n_components=20)
        lsi.fit(X_vect)
        X = lsi.transform_lsi_norm(X_vect)
    else:
        lsi = None
        X = X_vect

    s = Search(vect, idf, lsi)
    s.fit(X)

    for query, best_id in [(corpus[2], 2),
                           ('death dreams', 10)]:
        dist = s.search(query)
        assert dist.shape == (X.shape[0],)
        assert dist.argmax() == best_id
        # 2 - cosine distance should be in [0, 2]
        assert_array_less(dist, 1.001)
        assert_array_less(-1 - 1e-9, dist)



"""

To run this benchmark, you will need,

 * scikit-learn
 * pandas
 * memory_profiler
 * psutil (optional, but recommended)

"""

from __future__ import print_function

import numbers
import timeit
import itertools
import sys

from tqdm import tqdm
import numpy as np
import pandas as pd
from memory_profiler import memory_usage
import scipy.sparse as sp

from sklearn.datasets import fetch_20newsgroups
from sklearn.feature_extraction.text import (CountVectorizer, TfidfVectorizer,
                                             HashingVectorizer, strip_accents_ascii,
                                             strip_accents_unicode, defaultdict,
                                             _make_int_array)
from cytoolz import identity, compose, juxt, concat, frequencies, unique, count, merge
from cytoolz.curried import curry, sliding_window, map as cmap, remove, get_in
import toolz
from toolz.sandbox.core import unzip
from cytoolz.curried import keymap

n_repeat = 3


class FastCountVectorizer(CountVectorizer):
    """Convert a collection of text documents to a matrix of token counts
    This implementation produces a sparse representation of the counts using
    scipy.sparse.csr_matrix.
    If you do not provide an a-priori dictionary and you do not use an analyzer
    that does some kind of feature selection then the number of features will
    be equal to the vocabulary size found by analyzing the data.
    Read more in the :ref:`User Guide <text_feature_extraction>`.
    Parameters
    ----------
    input : string {'filename', 'file', 'content'}
        If 'filename', the sequence passed as an argument to fit is
        expected to be a list of filenames that need reading to fetch
        the raw content to analyze.
        If 'file', the sequence items must have a 'read' method (file-like
        object) that is called to fetch the bytes in memory.
        Otherwise the input is expected to be the sequence strings or
        bytes items are expected to be analyzed directly.
    encoding : string, 'utf-8' by default.
        If bytes or files are given to analyze, this encoding is used to
        decode.
    decode_error : {'strict', 'ignore', 'replace'}
        Instruction on what to do if a byte sequence is given to analyze that
        contains characters not of the given `encoding`. By default, it is
        'strict', meaning that a UnicodeDecodeError will be raised. Other
        values are 'ignore' and 'replace'.
    strip_accents : {'ascii', 'unicode', None}
        Remove accents during the preprocessing step.
        'ascii' is a fast method that only works on characters that have
        an direct ASCII mapping.
        'unicode' is a slightly slower method that works on any characters.
        None (default) does nothing.
    analyzer : string, {'word', 'char', 'char_wb'} or callable
        Whether the feature should be made of word or character n-grams.
        Option 'char_wb' creates character n-grams only from text inside
        word boundaries; n-grams at the edges of words are padded with space.
        If a callable is passed it is used to extract the sequence of features
        out of the raw, unprocessed input.
    preprocessor : callable or None (default)
        Override the preprocessing (string transformation) stage while
        preserving the tokenizing and n-grams generation steps.
    tokenizer : callable or None (default)
        Override the string tokenization step while preserving the
        preprocessing and n-grams generation steps.
        Only applies if ``analyzer == 'word'``.
    ngram_range : tuple (min_n, max_n)
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n
        will be used.
    stop_words : string {'english'}, list, or None (default)
        If 'english', a built-in stop word list for English is used.
        If a list, that list is assumed to contain stop words, all of which
        will be removed from the resulting tokens.
        Only applies if ``analyzer == 'word'``.
        If None, no stop words will be used. max_df can be set to a value
        in the range [0.7, 1.0) to automatically detect and filter stop
        words based on intra corpus document frequency of terms.
    lowercase : boolean, True by default
        Convert all characters to lowercase before tokenizing.
    token_pattern : string
        Regular expression denoting what constitutes a "token", only used
        if ``analyzer == 'word'``. The default regexp select tokens of 2
        or more alphanumeric characters (punctuation is completely ignored
        and always treated as a token separator).
    max_df : float in range [0.0, 1.0] or int, default=1.0
        When building the vocabulary ignore terms that have a document
        frequency strictly higher than the given threshold (corpus-specific
        stop words).
        If float, the parameter represents a proportion of documents, integer
        absolute counts.
        This parameter is ignored if vocabulary is not None.
    min_df : float in range [0.0, 1.0] or int, default=1
        When building the vocabulary ignore terms that have a document
        frequency strictly lower than the given threshold. This value is also
        called cut-off in the literature.
        If float, the parameter represents a proportion of documents, integer
        absolute counts.
        This parameter is ignored if vocabulary is not None.
    max_features : int or None, default=None
        If not None, build a vocabulary that only consider the top
        max_features ordered by term frequency across the corpus.
        This parameter is ignored if vocabulary is not None.
    vocabulary : Mapping or iterable, optional
        Either a Mapping (e.g., a dict) where keys are terms and values are
        indices in the feature matrix, or an iterable over terms. If not
        given, a vocabulary is determined from the input documents. Indices
        in the mapping should not be repeated and should not have any gap
        between 0 and the largest index.
    parallel : {bool, str}, default=False
        use parallel dask computations when needed
    binary : boolean, default=False
        If True, all non zero counts are set to 1. This is useful for discrete
        probabilistic models that model binary events rather than integer
        counts.
    dtype : type, optional
        Type of the matrix returned by fit_transform() or transform().
    Attributes
    ----------
    vocabulary_ : dict
        A mapping of terms to feature indices.
    stop_words_ : set
        Terms that were ignored because they either:
          - occurred in too many documents (`max_df`)
          - occurred in too few documents (`min_df`)
          - were cut off by feature selection (`max_features`).
        This is only available if no vocabulary was given.
    See also
    --------
    HashingVectorizer, TfidfVectorizer
    Notes
    -----
    The ``stop_words_`` attribute can get large and increase the model size
    when pickling. This attribute is provided only for introspection and can
    be safely removed using delattr or set to None before pickling.
    """

    def __init__(self, input='content', encoding='utf-8',
                 decode_error='strict', strip_accents=None,
                 lowercase=True, preprocessor=None, tokenizer=None,
                 stop_words=None, token_pattern=r"(?u)\b\w\w+\b",
                 ngram_range=(1, 1), analyzer='word',
                 max_df=1.0, min_df=1, max_features=None,
                 vocabulary=None, binary=False, dtype=np.int64,
                 parallel=True):
        self.input = input
        self.encoding = encoding
        self.decode_error = decode_error
        self.strip_accents = strip_accents
        self.preprocessor = preprocessor
        self.tokenizer = tokenizer
        self.analyzer = analyzer
        self.lowercase = lowercase
        self.token_pattern = token_pattern
        self.stop_words = stop_words
        self.max_df = max_df
        self.min_df = min_df
        if max_df < 0 or min_df < 0:
            raise ValueError("negative value for max_df or min_df")
        self.max_features = max_features
        if max_features is not None:
            if (not isinstance(max_features, numbers.Integral) or
                    max_features <= 0):
                raise ValueError(
                    "max_features=%r, neither a positive integer nor None"
                    % max_features)
        self.ngram_range = ngram_range
        self.vocabulary = vocabulary
        self.binary = binary
        self.dtype = dtype
        self.parallel = parallel

    def build_preprocessor(self):
        """Return a function to preprocess the text before tokenization"""
        if self.preprocessor is not None:
            return self.preprocessor

        # unfortunately python functools package does not have an efficient
        # `compose` function that would have allowed us to chain a dynamic
        # number of functions. However the cost of a lambda call is a few
        # hundreds of nanoseconds which is negligible when compared to the
        # cost of tokenizing a string of 1000 chars for instance.

        # accent stripping
        if not self.strip_accents:
            strip_accents = identity
        elif callable(self.strip_accents):
            strip_accents = self.strip_accents
        elif self.strip_accents == 'ascii':
            strip_accents = strip_accents_ascii
        elif self.strip_accents == 'unicode':
            strip_accents = strip_accents_unicode
        else:
            raise ValueError('Invalid value for "strip_accents": %s' %
                             self.strip_accents)

        if self.lowercase:
            return compose(strip_accents, str.lower)
        else:
            return strip_accents

    def build_analyzer(self):
        """Return a callable that handles preprocessing and tokenization"""
        if callable(self.analyzer):
            return self.analyzer

        preprocess = self.build_preprocessor()

        if self.analyzer == 'char':
            return compose(self._char_ngrams, preprocess, self.decode)

        elif self.analyzer == 'char_wb':
            return compose(self._char_wb_ngrams, preprocess, self.decode)

        elif self.analyzer == 'word':
            stop_words = self.get_stop_words()
            tokenize = self.build_tokenizer()
            ngram_builder = self._word_ngrams(stop_words=stop_words)

            return compose(ngram_builder, tokenize, preprocess, self.decode)

        else:
            raise ValueError('%s is not a valid tokenization scheme/analyzer' %
                             self.analyzer)

    def _word_ngrams(self, stop_words=None):
        """Turn tokens into a sequence of n-grams after stop words filtering"""
        # handle stop words
        if stop_words is not None:
            stop_words_filter = remove(lambda x: x in stop_words)
        else:
            stop_words_filter = identity

        # handle token n-grams
        min_n, max_n = self.ngram_range
        if min_n == max_n == 1:
            return stop_words_filter
        else:
            return compose(concat,
                           juxt([compose(cmap(' '.join), sliding_window(ngram))
                                 for ngram in range(min_n, max_n+1)]),
                           stop_words_filter)

    def fit(self, raw_documents, y=None):
        """Learn a vocabulary dictionary of all tokens in the raw documents.
        Parameters
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode or file objects.
        Returns
        -------
        self
        """
        analyze = self.build_analyzer()
        if not self.parallel:
            res = list(compose(unique, concat, cmap(compose(unique, analyze)))(raw_documents))
        else:
            import dask.bag as db
            # version 2 => 2.38, 5.78
            seq = db.from_sequence(raw_documents, npartitions=5).map(analyze)
            def reduce_partitions(vectors):
                return set(el for row in vectors for el in row)
            res = seq.map_partitions(reduce_partitions).distinct().compute()

            # version 3 => 2.59, 26.64
            #seq = db.from_sequence(raw_documents, npartitions=5).map(compose(set, analyze))
            #res = sorted(list(seq.fold(set.union).compute()))

            #print(res)

        return self

    def _count_vocab(self, raw_documents, fixed_vocab):
        """Create sparse feature matrix, and vocabulary where fixed_vocab=False
        """
        vocabulary = {'test': 0, 'this': 1, 'etc': 2}
        #if fixed_vocab:
        #    vocabulary = self.vocabulary_
        #else:
        #    # Add a new value when a new vocabulary item is seen
        #    vocabulary = defaultdict()
        #    vocabulary.default_factory = vocabulary.__len__

        #analyze = self.build_analyzer()
        #j_indices = []
        #indptr = _make_int_array()
        #values = _make_int_array()
        #indptr.append(0)
        #for doc in raw_documents:
        #    feature_counter = {}
        #    for feature in analyze(doc):
        #        try:
        #            feature_idx = vocabulary[feature]
        #            if feature_idx not in feature_counter:
        #                feature_counter[feature_idx] = 1
        #            else:
        #                feature_counter[feature_idx] += 1
        #        except KeyError:
        #            # Ignore out-of-vocabulary items for fixed_vocab=True
        #            continue

        #    j_indices.extend(feature_counter.keys())
        #    values.extend(feature_counter.values())
        #    indptr.append(len(j_indices))

        #if not fixed_vocab:
        #    # disable defaultdict behaviour
        #    vocabulary = dict(vocabulary)
        #    if not vocabulary:
        #        raise ValueError("empty vocabulary; perhaps the documents only"
        #                         " contain stop words")

        #j_indices = np.asarray(j_indices, dtype=np.intc)
        #indptr = np.frombuffer(indptr, dtype=np.intc)
        #values = np.frombuffer(values, dtype=np.intc)

        #X = sp.csr_matrix((values, j_indices, indptr),
        #                  shape=(len(indptr) - 1, len(vocabulary)),
        #                  dtype=self.dtype)
        #X.sort_indices()

        analyze = self.build_analyzer()
        res = list(map(compose(frequencies, analyze), raw_documents))
        vocab = {key: idx for idx, key in enumerate(sorted(unique(concat(el.keys() for el in res))))}
        def token2id(token):
            return vocab[token]
        res2 = map(keymap(token2id), res)
        res3 = unzip((el.keys(), el.values()) for el in res2)
        
        indices = np.asarray(list(concat(res3[0])))
        data = np.asarray(list(concat(res3[1])))
        #print(indices)
        #print(vocab)
        X = sp.csr_matrix(np.ones((1, 3)))
        return vocabulary, X


vect = FastCountVectorizer(ngram_range=(1, 2))

analyzer = vect.build_analyzer()

#print(analyzer('This document thing'))

#sys.exit()


def run_vectorizer(Vectorizer, X, **params):
    def f():
        vect = Vectorizer(**params)
        vect.fit(X)
    return f


text = fetch_20newsgroups(subset='train').data#[:2000]

print("="*80 + '\n#' + "    Text vectorizers benchmark" + '\n' + '='*80 + '\n')
print("Using a subset of the 20 newsrgoups dataset ({} documents)."
      .format(len(text)))
print("This benchmarks runs in ~20 min ...")

res = []

for Vectorizer, (analyzer, ngram_range) in tqdm(itertools.product(
            [FastCountVectorizer, CountVectorizer],
            [('word', (1, 1)),
             #('word', (1, 2)),
             #('word', (1, 4)),
             #('char', (4, 4)),
             #('char_wb', (4, 4))
             ])):

    bench = {'vectorizer': Vectorizer.__name__}
    params = {'analyzer': analyzer, 'ngram_range': ngram_range}
    bench.update(params)
    dt = timeit.repeat(run_vectorizer(Vectorizer, text, **params),
                       number=1,
                       repeat=n_repeat)
    bench['time'] = "{:.2f} (+-{:.2f})".format(np.mean(dt), np.std(dt))

    mem_usage = memory_usage(run_vectorizer(Vectorizer, text, **params))

    bench['memory'] = "{:.1f}".format(np.max(mem_usage))

    res.append(bench)


df = pd.DataFrame(res).set_index(['analyzer', 'ngram_range', 'vectorizer'])

print('\n========== Run time performance (sec) ===========\n')
print('Computing the mean and the standard deviation '
      'of the run time over {} runs...\n'.format(n_repeat))
print(df['time'].unstack(level=-1))

print('\n=============== Memory usage (MB) ===============\n')
print(df['memory'].unstack(level=-1))



# -*- coding: utf-8 -*-

import os
import warnings
import re
from collections import Counter

import numpy as np
import pandas as pd

from freediscovery.exceptions import (NotFound, WrongParameter)


def _list_filenames(data_dir, dir_pattern=None, file_pattern=None):
    """ List all files in a data_dir"""
    import re
    # parse all files in the folder
    filenames = []
    for root, subdirs, files in os.walk(data_dir):
        if dir_pattern is None or re.match(dir_pattern, root):
            for fname in files:
                if file_pattern is None or re.match(file_pattern, fname):
                    filenames.append(os.path.normpath(os.path.join(root,
                                                                   fname)))

    # make sure that sorting order is deterministic
    return sorted(filenames)


def _infer_document_id_from_path(file_path):
    basename = os.path.basename
    document_id = [re.sub('\D', '', basename(el)) for el in file_path]
    non_digits = [el for el in document_id if not el.isdigit()]
    failed_msg = ('Warning: Could not infer document_id from file_path ({}), '
                  'falling back '
                  'to `document_id_generator="indexed_file_path"`.')
    if non_digits:
        print(failed_msg.format('{} file_path do not contain any digits'
                                .format(len(non_digits))))
        return None
    document_id_counts = [(item, count)
                          for item, count in Counter(document_id).items()
                          if count > 1]
    if document_id_counts:
        print(failed_msg.format('{} document_id inferred from file_path '
                                'are non uniques'
                                .format(document_id_counts)))
        return None

    return np.array(document_id, dtype=int)


def _generate_document_id(db, document_id_generator):
    if 'document_id' not in db:
        if document_id_generator == 'infer_file_path':
            document_id = _infer_document_id_from_path(db.file_path)
            if document_id is None:
                # infering failed
                document_id_generator = 'indexed_file_path'
            else:
                db['document_id'] = document_id

        if document_id_generator == 'indexed_file_path':
            db['document_id'] = db.internal_id


def _check_mutual_index(keys1, keys2):
    """ Given two datasets with columns keys1 and keys2,
    returns the columns that could be used as an index
    """
    if 'internal_id' in keys1 and 'internal_id' in keys2:
        index_cols = ['internal_id']
    elif "document_id" in keys1 and \
         "document_id" in keys2 and \
         "rendition_id" in keys1 and \
         "rendition_id" in keys2:
        index_cols = ['document_id', 'rendition_id']
    elif "document_id" in keys1 and \
         "document_id" in keys2:
        index_cols = ['document_id']
    elif "file_path" in keys1 and \
         "file_path" in keys2:
        index_cols = ['file_path']
    else:
        raise ValueError("Cannot create a mutual index from "
                         "columns\n keys1 : {}\n keys2: {}".format(
                               list(keys1), list(keys2)))

    return index_cols


class DocumentIndex(object):
    def __init__(self, data_dir, data):
        self.data_dir = data_dir
        self.data = data

    def _check_index(self, keys=None):
        """ Given a list of keys check which keys will be used for indexing
        and whether these keys could be used for an index

        Parameters
        ----------
        keys : list
          one or multiple choices among "internal_id", "document_id",
          "rendition_id", "file_path".  default=["internal_id"]
        Returns
        -------
        index_cols : list
          a subset of keys that would be used for an index
        """

        if keys is None:
            keys = ['internal_id']

        index_cols = _check_mutual_index(keys, self.data.columns)
        if index_cols == ["document_id"] \
           and not self.data.document_id.is_unique:
            raise ValueError('document_id cannot be used as an index, '
                             'since it has duplicates '
                             ' (and rendition_id has duplicates)')

        if len(index_cols) == 1:
            index_cols = index_cols[0]

        # make sure we can use the selected columns as an index
        self.data.set_index(index_cols, verify_integrity=True)
        return index_cols

    def search(self, query, strict=True, drop=True,
               return_file_path=False):
        """Search the filenames given by some user query

        Parameters
        ----------
        query : pandas.DataFrame
           a DataFrame with one of the following fields "internal_id",
           ("document_id", "rendition_id"), "document_id", "file_path"
        strict : bool
           raise an error if some documents are not found
        drop : bool
           drop columns not in the dataset

        Returns
        -------
        df : pd.DataFrame
            the response dataframe with fields
            "internal_id", "file_path" and optionally "document_id"
            and "rendition_id"
        """
        if not isinstance(query, pd.DataFrame):
            raise ValueError('The query {} must be a pandas DataFrame')
        if not query.shape[0]:
            raise ValueError('Query has zero element!')

        if 'file_path' not in self.data.columns:
            if 'file_path' in query.columns or return_file_path:
                self.data['file_path'] = self.filenames_

        index_cols = self._check_index(query.columns)

        query['sort_order'] = query.index.values

        res = self.data.merge(query, on=index_cols, how='inner',
                              suffixes=('', '_query'))
        # make sure we preserve the original order in the query
        res.sort_values(by='sort_order', inplace=True)
        del res['sort_order']

        if res.shape[0] != query.shape[0]:
            # some documents were not found
            msg = ['{} out of {} query elements not found'
                   '\n(using "{}" as index):'.format(
                          query.shape[0] - res.shape[0],
                          query.shape[0], index_cols)]
            for index, row in query.iterrows():
                if row[index_cols] not in self.data[index_cols].values:
                    msg.append('   * {}'.format(row.to_dict()))

            msg.append('Expected format: \n {}'.format(
                       self.data.iloc[0, :].to_dict()))
            msg.append('with a total of {} documents'.format(
                       self.data.shape[0]))

            if strict:
                raise NotFound('\n'.join(msg))
            else:
                print('Warning: ' + '\n'.join(msg))

        if drop:
            # ignore all additional columns
            res = res[self.data.columns]

        return res

    def _search_filenames(self, filenames):
        """ A helper function that reproduces the previous
        behaviour in FeaturesVectorizer"""
        query = pd.DataFrame(filenames, columns=['file_path'])

        res = self.search(query)
        return res.internal_id.values

    def render_dict(self, res=None, return_file_path=False):
        """Render a pandas dataframe as a list of dicts

        Parameters
        ----------
        res : {pandas.DataFrame, None}
            some dataset with additional data that must contain
            the 'internal_id' key
        return_file_path : bool
            return the file paths, default: False

        Results
        -------
        out : dict

        """
        if res is not None:
            res = res.set_index('internal_id', drop=False)
        db = self.data.set_index('internal_id', drop=False)
        if not return_file_path:
            base_keys = [key for key in self.data.columns
                         if key != 'file_path']
        else:
            base_keys = list(self.data.columns)
            if 'file_path' not in db.columns:
                db['file_path'] = self.filenames_
                base_keys += ['file_path']

        if res is not None:
            res_keys = [key for key in res
                        if key not in base_keys]
            if not return_file_path and 'file_path' in res_keys:
                res_keys.remove('file_path')

        db = db[base_keys]

        if res is not None:
            db_dict = res[res_keys].join(db, how='left')
            out = db_dict.to_dict(orient='records')
        else:
            out = db.to_dict(orient='records')

        return out

    def render_list(self, res=None, return_file_path=False):
        """Render a pandas dataframe as a dict of lists

        Parameters
        ----------
        res : {pandas.DataFrame, None}
            some dataset with additional data that must contain
            the 'internal_id' key
        return_file_path : bool
            return the file paths, default: False

        Results
        -------
        out : dict

        """
        # TODO: this can be better rewritten with df.to_dict()
        if res is not None:
            res = res.set_index('internal_id', drop=False)
        db = self.data.set_index('internal_id', drop=False)
        if not return_file_path:
            base_keys = [key for key in self.data.columns
                         if key != 'file_path']
        else:
            base_keys = list(self.data.columns)
        db = db[base_keys]

        if res is not None:
            res_keys = [key for key in res if key not in base_keys]
            if not return_file_path:
                if 'file_path' in res_keys:
                    res_keys.remove('file_path')
        else:
            res_keys = []

        out = {}
        for key in base_keys + res_keys:
            out[key] = []
        if res is not None:
            for index, row in res[res_keys].iterrows():
                db_sel_dict = db.loc[index].to_dict()
                for key, val in db_sel_dict.items():
                    out[key].append(val)
                for key, val in row.to_dict().items():
                    out[key].append(val)
        else:
            for index, row in db.iterrows():
                row_dict = row.to_dict()
                for key, val in row_dict.items():
                    out[key].append(val)
        return out

    @classmethod
    def from_list(cls, metadata, data_dir=None, internal_id_offset=0,
                  dsid_dir=None, document_id_generator='indexed_file_path'):
        """ Create a DocumentIndex from a list of dictionaries, for instance

        .. code:: javascript

            {
                document_id: 1,
                rendition_id: 4,
                file_path: "c:\dev\1.txt"
            }

        Parmaters
        ---------
        metadata : list of dicts
            a list of dictionaries with keys
            ['file_path', 'document_id', 'rendition_id']
            describing the data ingestion (this overwrites data_dir)
        data_dir : str
            path to an exisitng data dir
        internal_id_offset : int
            the offset for internal_id (used when appending files
            to an existing dataset)
        dsid_dir : str
            In case content is provided, store it to a raw folder
            in this directory

        Returns
        -------
        result : DocumentIndex
            a DocumentIndex object
        """
        has_file_path = np.array(['file_path' in row for row in metadata])
        has_content = np.array(['content' in row for row in metadata])
        if has_file_path.all():
            if has_content.any():
                warnings.warn("Both 'file_path' and 'content' fields are "
                              "provided, the latter will be ingored!")
            metadata = sorted(metadata, key=lambda x: x['file_path'])

            # modify the metadata list inplace
            for idx, db_el in enumerate(metadata):
                db_el['internal_id'] = idx + internal_id_offset
                db_el['file_path'] = os.path.normpath(db_el['file_path'])
        elif has_content.all():
            if not (dsid_dir / 'raw').exists():
                (dsid_dir / 'raw').mkdir()
            data_dir = str(dsid_dir / 'raw')
            for idx, db_el in enumerate(metadata):
                db_el['internal_id'] = idx + internal_id_offset
                if 'rendering_id' in db_el:
                    rendering_id_el = db_el['rendering_id']
                else:
                    rendering_id_el = 0
                file_path_el = '{:09}_{}.txt'.format(idx + internal_id_offset,
                                                     rendering_id_el)
                db_el['file_path'] = os.path.join(data_dir, file_path_el)
                with (dsid_dir / 'raw' / file_path_el).open('wt', encoding='utf-8') as fh:
                    fh.write(db_el.pop('content'))

        elif (~has_file_path).any():
            raise ValueError(("All the ingested elements must have a "
                              "'file_path' or 'content' field. Currently, \n"
                              " - {} / {} documents have the 'file_path' "
                              "field,\n"
                              " - {} / {} documents have the 'content' field."
                              ).format(has_file_path.sum(), len(has_file_path),
                                       has_content.sum(), len(has_content)))

        filenames = [el['file_path'] for el in metadata]
        db = pd.DataFrame(metadata)

        _generate_document_id(db, document_id_generator)

        res = cls(data_dir, db)
        res.filenames_ = filenames
        return res

    @staticmethod
    def _detect_data_dir(filenames):
        if len(filenames) > 1:
            data_dir = os.path.commonprefix(filenames)
            data_dir = os.path.normpath(data_dir)
        elif len(filenames) == 1:
            data_dir = os.path.dirname(filenames[0])
        else:
            raise ValueError('Provided list of filenames is empty!')

        if os.path.exists(data_dir):
            return data_dir
        elif os.path.exists(os.path.dirname(data_dir)):
            return os.path.dirname(data_dir)
        else:
            raise IOError('data_dir={} does not exist!'.format(data_dir))

    @classmethod
    def from_folder(cls, data_dir, file_pattern=None, dir_pattern=None,
                    internal_id_offset=0, document_id_generator='indexed_file_path'):
        """ Create a DocumentIndex from files in data_dir

        Parmaters
        ---------
        data_dir : str
            path to the data directory (used only if metadata not provided),
            default: None

        Returns
        -------
        result : DocumentIndex
            a DocumentIndex object
        """

        data_dir = os.path.abspath(os.path.normpath(data_dir))

        if not os.path.exists(data_dir):
            raise NotFound('data_dir={} does not exist'.format(data_dir))

        if document_id_generator not in ['indexed_file_path', 'infer_file_path']:
            raise WrongParameter(("document_id_generator={} not supported. It must be "
                                  "one of ['indexed_file_path', 'infer_file_path']")
                                 .format(document_id_generator))

        filenames = _list_filenames(data_dir, dir_pattern, file_pattern)

        db = [{'file_path': file_path, 'internal_id': idx + internal_id_offset}
              for idx, file_path in enumerate(filenames)]

        db = pd.DataFrame(db)

        _generate_document_id(db, document_id_generator)

        res = cls(data_dir, db)
        res.filenames_ = filenames
        return res

    def _make_relative_paths(self):
        """ By default DocumentIndex uses absolute file paths,
        this makes the file_path to be relative to dir_path"""
        if self.data_dir is None:
            self.data_dir = self._detect_data_dir(self.filenames_)

        self.filenames_ = [os.path.relpath(el, self.data_dir)
                           for el in self.filenames_]
        self.data['file_path'] = self.filenames_



# -*- coding: utf-8 -*-
#
# Authors: Roman Yurchak
#
# License: BSD 3 clause

import numpy as np
from sklearn.base import BaseEstimator
from freediscovery.cluster.utils import _merge_clusters
from sklearn.utils.validation import check_array


class IMatchNearDuplicates(BaseEstimator):
    """
    Near duplicates detection using the randomized I-Match algorithm.

    A classical near-duplicates detection involves comparing all pairs of
    samples in the collection. For a collection of size ``N``, this is
    typically an ``O(N^2)`` operation. The I-Match algorithm allows to
    retrieve near duplicates with a computational effort reduced to
    ``O(N)`` (or ``O(N*log(N))`` in worse case scenario).

    This class exposes a scikit-learn compatible API, and currently supports
    only sparse CSR arrays (such as obtained after vectorizing text documents).


    Parameters
    ----------
     - n_rand_lexicons : int, default=1
        number of random lexicons used for duplicate detection
        If equal to 1 no lexicon randomization is used which is equivalent
        to the original I-Match implementation by Chowdhury et al. (2002).
     - rand_lexicon_ratio : float, default=0.7
        fraction of the vocabulary used in random lexicons.

    References
    ----------
    .. [Chowdhury2002]  `Chowdhury, A., Frieder, O., Grossman, D.,
                        & McCabe, M. C. (2002). Collection statistics for fast
                        duplicate document detection.
                        ACM Transactions on Information Systems (TOIS),
                        20(2), 171-191.`
    .. [Kolcz2008] - `Kołcz, A., & Chowdhury, A. (2008). Lexicon randomization
                      for near-duplicate detection with I-Match.
                      The Journal of Supercomputing, 45(3), 255-276.`
    """
    def __init__(self, n_rand_lexicons=1, rand_lexicon_ratio=0.7):
        self._fit_X = None
        self.n_rand_lexicons = n_rand_lexicons
        self.rand_lexicon_ratio = rand_lexicon_ratio
        if n_rand_lexicons < 1:
            raise ValueError
        if not (0 < rand_lexicon_ratio < 1):
            raise ValueError

    def fit(self, X, y=None):
        """
        Parameters
        ----------
        X : array_like or sparse (CSR) matrix, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        Returns
        -------
        self : object
            Returns self.
        """
        from sklearn.utils.murmurhash import murmurhash3_bytes_u32
        self._fit_X = X = check_array(X, accept_sparse='csr')

        n_samples, n_features = X.shape

        slice_list = [slice(None)]  # no lexicon randomization

        if self.n_rand_lexicons > 1:  # use lexicon randomization
            for _ in range(self.n_rand_lexicons - 1):
                # make a random choice of features that will be used
                slice_list.append(np.random.choice(n_features,
                                  int(self.rand_lexicon_ratio*n_features),
                                  replace=False))

        ihash_all = []
        for islice in slice_list:
            X_sl = X[:, islice]
            ihash = []
            for irow in range(n_samples):
                # compute features hash, don't use a random seed
                hash_res = murmurhash3_bytes_u32(X_sl[irow].indices.tobytes(),
                                                 seed=0)
                ihash.append(hash_res)
            ihash_all.append(ihash)

        # an array (n_samples, n_lexicons) with the document hashes for
        # each lexicon
        self.hash_ = np.array(ihash_all).T

        unique_gen = (np.unique(col, return_counts=True, return_inverse=True)
                      for col in self.hash_.T)
        # this is an integer array (n_samples, n_lexicons) with the size
        # of the cluster for each element
        # subsequently (self.hash_is_dup_ > 1).sum(axis=1) yields the
        # numeber of lexicons that suggest each document is a duplicate
        self.hash_is_dup_ = np.array([
                   counts[indices] for _, indices, counts in unique_gen]).T

        self.labels_ = _merge_clusters(self.hash_, rename=True)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from freediscovery.server import fd_app

app = fd_app('/freediscovery_shared/')



"""
Email threading
===============

An example illustrating the use of email threading algorithm
on the fedora mailing list.
"""
from __future__ import print_function

from time import time
import sys
import platform

import pandas as pd
import requests

pd.options.display.float_format = '{:,.3f}'.format


if platform.system() == 'Windows' and sys.version_info > (3, 0):
    print('This example currently fails on Windows with PY3 (issue #')
    sys.exit()

dataset_name = "fedora_ml_3k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL
###############################################################################
#
# 0. Load the test dataset
# -------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
res = requests.get(url)
res = res.json()

# To use a custom dataset, simply specify the following variables
data_dir = res['metadata']['data_dir']


###############################################################################
#
# 1. Parse emails
# ----------------

url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url, json={'parse_email_headers': True}).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))


url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
requests.post(url, json={'data_dir': data_dir})


###############################################################################
#
# 2. Thread Emails
# ----------------

url = BASE_URL + '/email-threading/'
print(" POST", url)
t0 = time()
res = requests.post(url, json={'parent_id': dsid}).json()

mid = res['id']
print("     => model id = {}".format(mid))


def print_thread(container, depth=0):
    print(''.join(['> ' * depth,  container['subject'],
                   ' (id={})'.format(container['id'])]))

    for child in container['children']:
        print_thread(child, depth + 1)

###############################################################################
#
# Threading examples
# cf. https://www.redhat.com/archives/rhl-devel-list/2008-October/thread.htlm
# for ground truth data (mailman has a maximum threading depth of 3,
# unlike FreeDiscovery


for idx in [-1, -2, -3, -4, -5]:  # get latest threads
    print(' ')
    print_thread(res['data'][idx])



# -*- coding: utf-8 -*-

import pytest
from unittest import SkipTest

from freediscovery.utils import dict2type

from .base import parse_res, V01, app, app_notest, get_features_cached

# ============================================================================#
#
#                     Duplicates detection
#
# ============================================================================#


@pytest.mark.parametrize('kind, options', [['simhash', {'distance': 3}],
                                           ['i-match', {}]])
def test_api_dupdetection(app, kind, options):

    if kind == 'simhash':
        try:
            import simhash
        except ImportError:
            raise SkipTest

    dsid, pars, _ = get_features_cached(app, hashed=False)

    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)

    url = V01 + "/duplicate-detection"
    pars = {'parent_id': dsid,
            'method': kind}
    data = app.post_check(url, json=pars)
    assert dict2type(data) == {'id': 'str'}
    mid = data['id']

    url += '/{}'.format(mid)
    data = app.get_check(url, query_string=options)

    assert dict2type(data, max_depth=1) == {'data': 'list'}
    for row in data['data']:
        assert dict2type(row, max_depth=1) == {'cluster_id': 'int',
                                               'cluster_similarity': 'float',
                                               'documents': 'list'}

    app.delete_check(url)



# -*- coding: utf-8 -*-

import os.path

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.email_threading import _EmailThreadingWrapper
from freediscovery.tests.run_suite import check_cache


basename = os.path.dirname(__file__)
data_dir = os.path.join(basename, "..", "..", "data",
                        "fedora-devel-list-2008-October")


def test_threading():
    cache_dir = check_cache()

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    uuid = fe.setup()
    fe.ingest(data_dir=data_dir)
    fe.parse_email_headers()

    cat = _EmailThreadingWrapper(cache_dir=cache_dir, parent_id=uuid)

    tree = cat.thread()
    cat.get_params()

    tree_ref = [{'id': 0, 'parent': None, 'children': [
                {'id': 1, 'children': [], 'parent': 0},
                {'id': 2, 'parent': 0,  'children': [
                         {'id': 3, 'children': [], 'parent': 2},
                         {'id': 4, 'children': [], 'parent': 2}]}]}]

    assert [el.to_dict() for el in tree] == tree_ref

    assert len(fe.filenames_) == sum([el.tree_size for el in tree])
    assert len(fe.filenames_) == 5
    assert len(tree[0].flatten()) == 5

    # res2 = {idx: [el0[0] for el0 in el] for idx, el in \
    #    groupby(enumerate(tree), key=lambda x: x[1])}

    # for key, idx in res2.items():
    #    print('Thread id: {}'.format(key))
    #    print('Relevant documents: {}'.format(idx))
    #    print('Parent documents  : {}'.format([parent[el] for el in idx]))



# -*- coding: utf-8 -*-

import os

from sklearn.externals import joblib

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.pipeline import PipelineFinder
from freediscovery.engine.utils import validate_mid
from freediscovery.exceptions import WrongParameter


class _BaseWrapper(object):
    """Base class for wrappers in FreeDiscovery

    Parameters
    ----------
    cache_dir : str
      folder where the model will be saved
    parent_id : str, optional
      dataset id
    mid : str, optional
      model id
    dataset_definition : a dataset defintion class
      one of FeatureVectorizer, EmailParser
    load_model : bool
      whether the model should be loaded from disk on class
      initialization
    mode : bool
      read/write mode. One of 'r', 'w', 'fw' (write, overwrite if exists)
    """
    def __init__(self, cache_dir='/tmp/', parent_id=None, mid=None,
                 load_model=False, mode='r'):
        if parent_id is None and mid is None:
            raise WrongParameter('At least one of parent_id or mid '
                                 'should be provided!')

        if self._wrapper_type == 'lsi' and self.mode in ['w', 'fw']:
            # lsi supports explicitly providing mid at creation
            if parent_id is None:
                raise WrongParameter(('parent_id={} must be provided for '
                                      'model creation!')
                                     .format(parent_id))
            else:
                validate_mid(parent_id)
                self.pipeline = PipelineFinder.by_id(parent_id, cache_dir)
                if mid is not None:
                    validate_mid(mid)
                self.mid = mid
        else:
            if parent_id is None and mid is not None:
                validate_mid(mid)
                self.pipeline = PipelineFinder.by_id(mid, cache_dir).parent
                self.mid = mid
            elif parent_id is not None:
                validate_mid(parent_id)
                self.pipeline = PipelineFinder.by_id(parent_id, cache_dir)
                self.mid = None

        # this only affects LSI
        if mode not in ['r', 'w', 'fw']:
            raise WrongParameter('mode={} must be one of "r", "w", "fw"'
                                 .format(mode))
        self.mode = mode

        # this is an alias that should be deprecated
        self.fe = FeatureVectorizer(cache_dir=cache_dir,
                                    dsid=self.pipeline['vectorizer'])

        self.model_dir = self.pipeline.get_path() / self._wrapper_type

        if self._wrapper_type == 'search':
            # no data need to be stored on disk
            return

        if not self.model_dir.exists():
            self.model_dir.mkdir()

        if self.mid is not None and self.mode == 'r':
            self._pars = self._load_pars()
        else:
            self._pars = None

        if load_model:
            if self.mid is not None and self.mode == 'r':
                self.cmod = self._load_model()
            else:
                self.cmod = None

    def delete(self):
        """ Delete a trained model"""
        import shutil
        mid_dir = self.model_dir / self.mid
        shutil.rmtree(str(mid_dir), ignore_errors=True)

    def __contains__(self, mid):
        return (self.model_dir / mid).exists()

    def get_params(self):
        """ Get model parameters """
        return self._pars

    def _load_model(self):
        """ Load model from disk """
        mid = self.mid
        mid_dir = self.model_dir / mid
        if not mid_dir.exists():
            raise ValueError('Model id {} ({}) not found in the cache {}!'
                             .format(mid, self._wrapper_type, mid_dir))
        cmod = joblib.load(str(mid_dir / 'model'))
        return cmod

    def _load_pars(self, mid=None):
        """Load model parameters from disk"""
        if mid is None:
            mid = self.mid
        model_path = self.model_dir / mid
        if not model_path.exists():
            raise ValueError('Model id {} ({}) not found in the cache {}!'
                             .format(mid, self._wrapper_type, model_path))
        pars = joblib.load(str(model_path / 'pars'))
        pars['id'] = mid
        return pars

    def list_models(self):
        """ List existing models of this type """
        out = []
        if not self.model_dir.exists():
            return out
        for mid in os.listdir(str(self.model_dir)):
            try:
                pars = self._load_pars(mid)
                pars['id'] = mid
                out.append(pars)
            except:
                raise
        return out




import sys

from flask_apispec import (marshal_with,
                           MethodResource as Resource)
from flask_apispec.annotations import doc
from webargs import fields as wfields

from webargs.core import argmap2schema
from .._version import __version__

# =========================================================================== #
#                        Server info                                          #
# =========================================================================== #


class ServerInfoApi(Resource):

    @doc(description="Return FreeDiscovery server information "
                     " (versions, etc).")
    @marshal_with(argmap2schema(
                  {'version': wfields.Nested({'number': wfields.Str()}),
                   'env': wfields.Nested({'python_version': wfields.Str()}),
                   'config': wfields.Nested({'cache_dir': wfields.Str(),
                                             'debug': wfields.Boolean(),
                                             'hostname': wfields.Str(),
                                             'log_file': wfields.Str(),
                                             'n_workers': wfields.Int(),
                                             'port': wfields.Int(),
                                             'server': wfields.Str()})
                   }))
    def get(self):
        out = {'version': {},
               'env': {}}

        out['version']['number'] = __version__
        out['env']['python_version'] = sys.version
        out['config'] = self._fd_config
        return out



"""
Hierarchical Clustering Example
===============================

Visualize hierarchical clusters
"""

import os.path
import pandas as pd
from time import time
import requests
from graphviz import Digraph


dataset_name = "20_newsgroups_3categories"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

###############################################################################
#
# 0. Load the example dataset
# ---------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()

# To use a custom dataset, simply specify the following variables
data_dir = input_ds['metadata']['data_dir']
dataset_definition = [{'document_id': row['document_id'],
                       'file_path': os.path.join(data_dir, row['file_path'])}
                      for row in input_ds['dataset']]

###############################################################################
#
# # 1. Feature extraction (non hashed)
# ------------------------------------
# 1.a Load dataset and initalize feature extraction

url = BASE_URL + '/feature-extraction'
print(" POST", url)
fe_opts = {'max_df': 0.6,  # filter out (too)/(un)frequent words
           'weighting': "ntc",
           }
res = requests.post(url, json=fe_opts).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))


###############################################################################
#
# 1.b Run feature extraction

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
res = requests.post(url, json={'dataset_definition': dataset_definition})

###############################################################################
#
# 2. Calculate LSI
# ----------------

url = BASE_URL + '/lsi/'
print("POST", url)

n_components = 100
res = requests.post(url,
                    json={'n_components': n_components,
                          'parent_id': dsid
                          }).json()

lsi_id = res['id']
print('  => LSI model id = {}'.format(lsi_id))
print(('  => SVD decomposition with {} dimensions '
       'explaining {:.2f} % variabilty of the data')
      .format(n_components, res['explained_variance']*100))



###############################################################################
#
# 3. Document Clustering (LSI + Birch Clustering)
# -----------------------------------------------
# 3.a. Document clustering (LSI + Birch clustering)

url = BASE_URL + '/clustering/birch/'
print(" POST", url)
t0 = time()
res = requests.post(url,
                    json={'parent_id': lsi_id,
                          'n_clusters': -1,
                          'min_similarity': 0.55,
                          #'max_tree_depth': 3,
                          }).json()

mid = res['id']
print("     => model id = {}".format(mid))

print("\n4.b. Computing cluster labels")
url = BASE_URL + '/clustering/birch/{}'.format(mid)
print(" GET", url)
res = requests.get(url,
                   json={'n_top_words': 3
                         }).json()
t1 = time()

print('    .. computed in {:.1f}s'.format(t1 - t0))
data = res['data']

print(pd.DataFrame(data))


###############################################################################
#
# 3.b Hierarchical cluster visualization

ch = Digraph('cluster_hierarchy',
             node_attr={'shape': 'record'},
             format='png')

ch.graph_attr['rankdir'] = 'LR'
ch.graph_attr['dpi'] = "200"

for row in data:
    ch.node('cluster_{}'.format(row['cluster_id']),
            '{{<f0>{}| {{<f1> id={:03}  |<f2> N={} |<f3> sim={:.2f} }}}}'
            .format(row['cluster_label'],
                    row['cluster_id'],
                    row['cluster_size'],
                    row['cluster_similarity']))


def create_hc_links(node, ch, data):
    for child_id in node['children']:
        ch.edge('cluster_{}:f2'.format(node['cluster_id']),
                'cluster_{}:f0'.format(child_id))
        create_hc_links(data[child_id], ch, data)


create_hc_links(data[0], ch, data)

tmp_dir = os.path.join('..', '..', 'doc', 'engine', 'examples')
if os.path.exists(tmp_dir):
    ch.render('cluster_hierarchy', directory=tmp_dir, cleanup=True)
else:
    ch.view()

####################################
# .. image:: cluster_hierarchy.png

###############################################################################
#
# 4. Delete the extracted features

url = BASE_URL + '/feature-extraction/{}'.format(dsid)
requests.delete(url)



# -*- coding: utf-8 -*-

from marshmallow import Schema as BaseSchema, fields


# This file defines marshmallow schemas for REST API output formatting
# Occasionally the same class can be used for input validation
# As webargs (used for input validation) is just a wrapper around marshmallow

class Schema(BaseSchema):

    class Meta:
        strict = True
        ordered = True


class DocumentIndexSchema(Schema):
    document_id = fields.Int()
    render_id = fields.Int()


class DocumentIndexFullSchema(DocumentIndexSchema):
    file_path = fields.Str()
    internal_id = fields.Int()


class _ExampleDatasetElementSchema(DocumentIndexFullSchema):
    category = fields.Str()


class _ExampleDatasetMetadataSchema(Schema):
    data_dir = fields.Str(required=True)
    name = fields.Str(required=True)


class ExampleDatasetSchema(Schema):
    metadata = fields.Nested(_ExampleDatasetMetadataSchema, required=True)
    training_set = fields.Nested(_ExampleDatasetElementSchema, many=True)
    dataset = fields.Nested(_ExampleDatasetElementSchema, many=True,
                            required=True)


class EmptySchema(Schema):
    pass


class IDSchema(Schema):
    id = fields.Str(required=True)


class DocumentIndexNestedSchema(Schema):
    data = fields.Nested(DocumentIndexFullSchema, many=True)


class _DatasetDefinition(Schema):
    document_id = fields.Int()
    rendition_id = fields.Int()
    file_path = fields.Str()
    content = fields.Str()


class _DatasetDefinitionShort(Schema):
    document_id = fields.Int()
    rendition_id = fields.Int()
    file_path = fields.Str()


class FeaturesParsSchema(Schema):
    data_dir = fields.Str()
    n_features = fields.Int(missing=100001)
    analyzer = fields.Str(missing='word')
    stop_words = fields.Str(missing='english')
    n_jobs = fields.Int(missing=1)
    chunk_size = fields.Int()
    ngram_range = fields.List(fields.Int(), missing=[1, 1])
    weighting = fields.Str(missing='nnc')
    norm_alpha = fields.Number(missing=0.75)
    use_hashing = fields.Boolean(missing=False)
    n_samples = fields.Int(dump_only=True)
    n_samples_processed = fields.Int(dump_only=True)
    min_df = fields.Number(required=False)
    max_df = fields.Number(required=False)
    parse_email_headers = fields.Boolean(missing=False)
    preprocess = fields.List(fields.Str(), missing=[])
    id = fields.Str()
    overwrite = fields.Boolean(missing=False)


class FeaturesSchema(FeaturesParsSchema):
    id = fields.Str(required=True)
    filenames = fields.List(fields.Str())


class ClassificationScoresSchema(Schema):
    precision = fields.Number(required=True)
    recall = fields.Number(required=True)
    f1 = fields.Number(required=True)
    roc_auc = fields.Number(required=True)
    average_precision = fields.Number(required=True)
    recall_at_20p = fields.Number()


class LsiParsSchema(Schema):
    parent_id = fields.Str(required=True)
    n_components = fields.Int(required=True)


class LsiPostSchema(IDSchema):
    explained_variance = fields.Number(required=True)


class _ResponsePaginaton(Schema):
    total_response_count = fields.Int()
    current_response_count = fields.Int()
    batch_id = fields.Int()
    batch_id_last = fields.Int()


class CategorizationPostSchema(Schema):
    id = fields.Str(required=True)
    training_scores = fields.Nested(ClassificationScoresSchema())


class _CategorizationIndex(DocumentIndexSchema):
    category = fields.Str(required=True)


class _CategorizationInputSchema(Schema):
    parent_id = fields.Str(required=True)
    # Warning this should be changed to wfields.DelimitedList
    # https://webargs.readthedocs.io/en/latest/api.html#webargs.fields.DelimitedList
    data = fields.Nested(_CategorizationIndex, many=True)
    method = fields.Str(default='LinearSVC')
    cv = fields.Boolean(missing=False)
    training_scores = fields.Boolean(missing=False)


class _CategorizationScoreSchemaElement(DocumentIndexSchema):
    category = fields.Str(required=True)
    score = fields.Number(required=True)


class _CategorizationPredictSchemaElement(DocumentIndexSchema):
    scores = fields.Nested(_CategorizationScoreSchemaElement, many=True,
                           required=True)


class CategorizationPredictSchema(Schema):
    data = fields.Nested(_CategorizationPredictSchemaElement, many=True,
                         required=True)
    pagination = fields.Nested(_ResponsePaginaton, required=True)


class CategorizationParsSchema(Schema):
    method = fields.Str(required=True)
    options = fields.Str(required=True)


class _ClusteringElementSubSchema(DocumentIndexSchema):
    similarity = fields.Number()


class _ClusteringElementSchema(Schema):
    cluster_id = fields.Int(required=True)
    cluster_similarity = fields.Number()
    cluster_label = fields.Str()
    cluster_depth = fields.Int()
    cluster_size = fields.Int()
    children = fields.List(fields.Int())
    documents = fields.Nested(_ClusteringElementSubSchema, many=True,
                              required=True)


class ClusteringSchema(Schema):
    data = fields.Nested(_ClusteringElementSchema, many=True, required=True)


class EmailThreadingParsSchema(Schema):
    group_by_subject = fields.Boolean(required=True)


class ErrorSchema(Schema):
    message = fields.Str(required=True)


class TreeSchema(Schema):
    id = fields.Int(required=True)
    parent = fields.Int(allow_none=True, required=True)
    subject = fields.Str()
    children = fields.Nested('self', many=True)


class EmailThreadingSchema(Schema):
    id = fields.Str(required=True)
    data = fields.Nested(TreeSchema, many=True)


class MetricsCategorizationSchema(Schema):
    precision = fields.Number()
    recall = fields.Number()
    f1 = fields.Number()
    roc_auc = fields.Number()
    average_precision = fields.Number()
    recall_at_20p = fields.Number()


class MetricsClusteringSchema(Schema):
    adjusted_rand = fields.Number()
    adjusted_mutual_info = fields.Number()
    v_measure = fields.Number()


class MetricsDupDetectionSchema(Schema):
    ratio_duplicates = fields.Number()
    f1_same_duplicates = fields.Number()
    mean_duplicates_count = fields.Number()


class _SearchResponseSchemaElement(DocumentIndexSchema):
    score = fields.Number(required=True)


class SearchResponseSchema(Schema):
    data = fields.Nested(_SearchResponseSchemaElement,
                         many=True, required=True)
    pagination = fields.Nested(_ResponsePaginaton, required=True)


class CustomStopWordsSchema(Schema):
    name = fields.Str(required=True)
    stop_words = fields.List(fields.Str(), required=True)


class CustomStopWordsLoadSchema(Schema):
    name = fields.Str(required=True)
    stop_words = fields.List(fields.Str(), required=True)



# Authors: Roman Yurchak
#
# License: BSD 3 clause

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array


class SimhashNearDuplicates(BaseEstimator):
    """Near duplicates detection using the simhash algorithm.

    A classical near-duplicates detection involves comparing all pairs of
    samples in the collection. For a collection of size ``N``, this is
    typically an ``O(N^2)`` operation. Simhash algorithm allows to
    retrieve near duplicates with a significantly better computational
    scaling.

    .. Note:: this estimator requires
              the `simhash-py <https://github.com/seomoz/simhash-py>`_Python package
              to be installed.

    Parameters
    ----------
    hash_func : str or function, default='murmurhash3_int_u32'
        the hashing function used to hash documents.
        Possibles values are "murmurhash3_int_u32" or a custom function.
    hash_func_nbytes : int, default=64
        expected size of the hash produced by hash_func

    References
    ----------
    .. [Charikar2002]  `Charikar, M. S. (2002, May).
                        Similarity estimation techniques from rounding
                        algorithms.
                        In Proceedings of the thiry-fourth annual ACM symposium
                        on Theory of computing (pp. 380-388). ACM.`
    """
    def __init__(self, hash_func='murmurhash3_int_u32', hash_func_nbytes=32):
        self._fit_X = None
        self._fit_shash_dict = {}
        if isinstance(hash_func, str):
            if hash_func == 'murmurhash3_int_u32':
                from sklearn.utils.murmurhash import murmurhash3_int_u32
                hash_func = murmurhash3_int_u32
                hash_func_nbytes = 32
            else:
                raise ValueError
        elif not hasattr(hash_func, '__call__'):
            raise ValueError

        self.hash_func = hash_func
        if hash_func_nbytes not in [32, 64]:
            raise ValueError('Hashing function other than 64bit '
                             'or 32bit are not supported!')

        self.hash_func_nbytes = hash_func_nbytes

    def fit(self, X, y=None):
        """
        Parameters
        ----------
        X : {array, sparse matrix}, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        Returns
        -------
        self : object
            Returns self.
        """
        from simhash import compute
        self._fit_X = X = check_array(X, accept_sparse='csr')

        n_features = X.shape[1]

        def _scale_hash_32_64bit(indices):
            return indices*((2**64-1)//2**32-1)

        hash_func = self.hash_func

        hashing_table = np.array(
                [hash_func(el, 0) for el in range(n_features)],
                dtype='uint64')

        shash = []
        for idx in range(X.shape[0]):
            # get hashes of indices
            mhash = hashing_table[X[idx].indices]
            if self.hash_func_nbytes == 32:
                mhash = _scale_hash_32_64bit(mhash)
            shash.append(compute(mhash))
        _fit_shash = np.asarray(shash, dtype='uint64')
        self._fit_shash = _fit_shash
        self._fit_shash_dict = {val: key
                                for key, val in enumerate(self._fit_shash)}

    def get_index_by_hash(self, shash):
        """ Get document index by hash

        Parameters
        ----------
        shash: uint64
           a simhash value

        Returns
        -------
        index: int
           a document index
        """
        return self._fit_shash_dict[shash]

    def query(self, distance=2, blocks='auto'):
        """ Find all the nearests neighbours for the dataset

        Parameters
        ----------
        distance : int, default=2
            Maximum number of differnet bits in the simhash
        blocks : int or 'auto', default='auto'
                number of blocks into which the simhash is split
                when searching for duplicates,
                see  https://github.com/seomoz/simhash-py

        Returns
        -------
        simhash : array
            the simhash value for all documents in the collection
        cluster_id : array
            the exact duplicates (documents with the same simhash)
            are grouped by in cluster_id
        dup_pairs : list
            list of tuples for the near-duplicates
        """
        from simhash import find_all

        if distance >= 64:
            raise ValueError(('Wrong input parameter for distance = {} '
                              'Must be less than 64!')
                             .format(distance))

        _, cluster_id = np.unique(self._fit_shash, return_inverse=True)

        if blocks == 'auto':
            blocks = min(distance*2, 64)
        matches = find_all(self._fit_shash, blocks, distance)
        matches = np.array(matches, dtype='uint64')
        return self._fit_shash, cluster_id, matches



# Authors: Denys Lazarenko
#          Roman Yurchak
#
# License: BSD 3 clause

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import numpy as np


def explain_categorization(weights, text, colormap):
    """Generate decorated HTML by adding <span> tags with background color to keywords depending on word's weight.

    Parameters
    ----------
    weights : dict[str, float]
        words as keys, weights (scores) as values
    text : str
        document's raw text
    colormap : matplotlib.colors.LinearSegmentedColormap
        color map used to decorate keywords with background color

    Returns
    -------
    str
        HTML string representing document's text decorated with <span> tags
    """
    try:
        from html import escape  # python 3.x
    except ImportError:
        from cgi import escape  # python 2.x
    from matplotlib.colors import Normalize

    text = escape(text)  # in order to have a valid HTML output after adding <span> tags

    max_val = max(abs(min(weights.values())), abs(max(weights.values())))

    norm = Normalize(vmin=-max_val, vmax=max_val)

    document_html_lines = list()
    for line in text.splitlines():
        line_decorated = _replace_with_color_spans(line, weights, colormap, norm)
        document_html_lines.append(line_decorated)
    return "<br/>".join(document_html_lines), norm


def _make_cmap(cmap_name='jet', alpha=0.2, filter_ratio=0.5):
    """Create a colormap which will be used to adding color spans to text

    Parameters
    ----------
    cmap_name : str
        name of colormap, see here for all possible values: http://matplotlib.org/users/colormaps.html
    alpha : float
        color's transparency
    filter_ratio : float
        make fully transparent (1 - filter_ratio) of the color bar

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap
        final color map with transparency
    """
    import matplotlib as mpl
    import matplotlib.cm as cm

    cmap = cm.get_cmap(cmap_name)

    # Extract colormap's colors and set new alpha
    cmap_array = cmap(np.arange(cmap.N))
    N = cmap_array.shape[0]
    if filter_ratio is not None:
        if not 0 <= filter_ratio <= 1:
            raise ValueError('filter_ratio = {} must be in the [0, 1] range'.format(filter_ratio))
        nf_ratio = 1 - filter_ratio
        cmap_array[:, -1] = 0.0
        cmap_array[:int(nf_ratio*N / 2), -1] = alpha
        cmap_array[-int(nf_ratio*N / 2):, -1] = alpha
    else:
        cmap_array[:, -1] = alpha


    # Create new colormap from the array with modified alpha
    cmap_with_trancparency = mpl.colors.ListedColormap(cmap_array)
    return cmap_with_trancparency


def _replace_with_color_spans(textline, weights, colormap, norm):
    """Given a line of text and a dictionary of word weights,
    add color span tag to those words in the textline that are present in the dictionary

    Parameters
    ----------
    textline : str
        text of one document's line
    weights : dict[str, float]
        words as keys, words' scores (weights) as values
    colormap : matplotlib.colors.LinearSegmentedColormap
        color map used to map the word's score to the background color of a <span>

    Returns
    -------
    str
        HTML string representing the word wrapped into a <span> tag with background color
    """

    html = textline
    positions = _get_keyword_positions(textline, weights.keys())
    positions_no_overlap = _keep_longest_overlapping_substrings(positions)


    # Perform replacement in the descreasing order of positions, i.e. from the end of the textline.
    # This way replacing words in the end of the string does not affect positions of words at the beginning
    for word_start, word_end in reversed(sorted(positions_no_overlap)):
        source_word = textline[word_start:word_end]
        # Using case-insensitive replacement while decorating with color spans.
        # Example:
        #   ERIC and Eric would be considered as the same word,
        # while keeping their original case.
        # We assume that dictionary has only lowercase words ('eric') in this case.
        key = source_word.lower()
        score = norm(weights[key])
        colored_word = _wrap_in_colored_span(source_word, score, colormap)
        # print(word_start, word_end, source_word, " ==> ", colored_word)
        html = html[:word_start] + colored_word + html[word_end:]
    return html


def _get_keyword_positions(textline, keywords):
    """Given a textline and a list of keywords, get a sorted list of positions of keywords found in the textline.

    Example
    -------
    textline: "Hello world, this is Eric."
    keywords: ["world", "eric"]
    would return [(6, 11), (21, 25)]. Each tuple represents word start and end index in the given string.

    Parameters
    ----------
    textline : str
        text of one document's line
    keywords : list[str]
        list of keywords for which we have scores (weights)

    Returns
    -------
    list[(int, int)]
        each tuple corresponds to a pair of (start, end) positions of a keyword inside a given line of text
    """
    positions_list = list()
    for word in keywords:
        if word in textline.lower():
            starts = [m.start() for m in re.finditer(word, textline.lower())]
            ends = [start+len(word) for start in starts]
            for start, end in zip(starts, ends):
                positions_list.append((start, end))
                # print(word, start, end)
    return sorted(positions_list)


def _keep_longest_overlapping_substrings(positions):
    """Deduplicate overlapping positions by keeping the longest possible keywords.

    Example
    -------
        document text contains string "email me at eric_bass@some-email.com"
        keywords vocabulary has "eric", "bass", "eric_bass"
        Thus keywords positions will have three tuples [(12, 16), (17, 21), (12, 21)].
        As we cannot overlap color spans, we have to deduplicate.
        In this examples we could either keep "eric" and "bass" or to keep "eric_bass" only and delete two others.
        We make a choice to keep the longest possible substring, deleting every other keyword overlapping with it.
        In this example the final output after deduplication will be [(12, 21].

    Example
    -------
        document text contains string "administrative procedures"
        keywords vocabulary has "administrative", "mini", "procedure", "pro"
        Thus keywords positions are [(0, 14), (2, 6), (15, 24), (15, 18)].
        Keeping longest possible keywords will eliminate "mini" and "pro" keywords,
        which were just accidentally matched and would represent a noise if kept.
        So, after deduplication only two tuples are left: [(0, 14), (15, 24)]

    Parameters
    ----------
    positions : list[(int, int)]
        each tuple corresponds to a pair of (start, end) positions of a keyword inside a given line of text

    Returns
    -------
    list[(int, int)]
        deduplicated version of keywords positions inside the text
    """
    overlap = True
    while overlap:
        overlap = False
        for idx in range(len(positions)-1):
            current_start, current_end = positions[idx]
            next_start, next_end = positions[idx+1]
            if current_end > next_start:
                overlap = True
                # In case of overlap, keep the longer substring,
                # and delete the indices of a shorter one from the list
                # Then break from the for-loop and restart analyzing the list
                if (current_end - current_start) > (next_end - next_start):
                    del positions[idx+1]
                else:
                    del positions[idx]
                break
    return sorted(positions)


def _wrap_in_colored_span(word, score, colormap):
    """Wrap a given word into an HTML <span> tag.

    Parameters
    ----------
    word : str
        word to be wrapped in a <span> tag with background color
    score : float
        word's score (weight)
    colormap : matplotlib.colors.LinearSegmentedColormap
        color map used to map the word's score to a background color of the span

    Returns
    -------
    str
        HTML tag <span> wrapped around the word to decorate it with a background color
    """
    rgba = _score_to_rgb(score, colormap)
    return '<span style="background-color: rgba({}, {}, {}, {});">{word}</span>'.format(*rgba, word=word)


def _score_to_rgb(score, colormap):
    """Convert word's score (weight) into an RGBa color.

    Parameters
    ----------
    score : float
        word's score (weight) from a dictionary
    colormap : matplotlib.colors.LinearSegmentedColormap
        color map used to map the word's score to an RGBa color

    Returns
    -------
    (int, int, int, float)
        RGBa representation of resulting color
    """
    mapped_rgba = colormap(score)
    r, g, b = [int(255*x) for x in list(mapped_rgba)[:3]]
    return r, g, b, mapped_rgba[-1]


def _create_random_weights(text, perc_keywords=0.5):
    """Create random weights for keywords.
    Keywords is a random subset of perc_keywords out of all words in the document.

    Parameters
    ----------
    text : str
        document text
    perc_keywords : float
        ratio of keywords of the whole vocabulary
    Returns
    -------
    dict[str: float]
        keywords as keys, their random weights as values
    """
    import numpy.random as rnd
    import random
    from sklearn.feature_extraction.text import CountVectorizer

    vect = CountVectorizer(input="content")
    _ = vect.fit([text])
    vocabulary = vect.vocabulary_
    nb_keywords = int(perc_keywords * len(vocabulary))
    keywords = random.sample(vocabulary.keys(), nb_keywords)
    features_weight = {word: rnd.random() for word in keywords}
    return features_weight


if __name__ == "__main__":
    fname = 'data/ds_001/raw/0.7.6.28635.txt'
    with open(fname) as in_file:  #, encoding='utf-8') as in_file:
        document_text = in_file.read().replace(u'\ufeff','')
    words_weights = _create_random_weights(document_text, 0.2)

    COLORMAP = _make_cmap()
    document_text_decorated = explain_categorization(words_weights, document_text, COLORMAP)
    with open('{}_decorated.html'.format(fname), 'w') as out_file:
        out_file.write(document_text_decorated)



# -*- coding: utf-8 -*-

from sklearn.externals import joblib
import numpy as np

from freediscovery.engine.base import _BaseWrapper
from freediscovery.utils import setup_model
from freediscovery.exceptions import WrongParameter
from freediscovery.engine.cluster import _BaseClusteringWrapper


class _DuplicateDetectionWrapper(_BaseWrapper, _BaseClusteringWrapper):
    """Find near duplicates in a document collection.

    Currently supported backends are simhash-py and i-match.

    The option `use_hashing=False` must be set for the feature extraction.
    Recommended options also include, `use_idf=1, sublinear_tf=0, binary=0`.

    Parameters
    ----------
    cache_dir : str
       directory where to save temporary and regression files
    parent_id : str, optional
       dataset id
    mid : str, optional
       model id
    """

    _wrapper_type = "dupdet"

    def __init__(self, cache_dir='/tmp/', parent_id=None, mid=None):

        super(_DuplicateDetectionWrapper, self).__init__(cache_dir=cache_dir,
                                                 parent_id=parent_id, mid=mid,
                                                 load_model=True)

        self.model = self.cmod
        del self.cmod

    def fit(self, method='simhash'):
        """
        Precompute all the required values for duplicate detection
        """

        pars = {'method': method}
        if method not in ['simhash', 'i-match']:
            raise WrongParameter('Dup. detection method {} not implemented!'.format(method))
        if method == 'simhash':
            from freediscovery.near_duplicates import SimhashNearDuplicates
            self.model = shash = SimhashNearDuplicates()
        else:
            self.model = None
        self._pars = pars
        mid, mid_dir = setup_model(self.model_dir)

        self.mid = mid

        X = self.pipeline.data
        if method == 'simhash':
            shash.fit(X)

        self._fit_X = X

        joblib.dump(self.model, str(self.model_dir / mid / 'model'))
        joblib.dump(pars, str(self.model_dir / mid / 'pars'))

    def query(self, **args):
        """ Find all the nearests neighbours for the dataset

        Parameters
        ----------
        distance : int, default=1
            Maximum number of differnet bits in the simhash
        blocks : int or 'auto', default='auto'
            number of blocks into which the simhash is split
            when searching for duplicates,
            see  https://github.com/seomoz/simhash-py
        Returns
        -------
        cluster_id : array
            the exact duplicates (documents with the same simhash)
            are grouped by in cluster_id
        """

        if self._pars['method'] == 'simhash':
            from freediscovery.cluster.utils import (_binary_linkage2clusters,
                                                     _merge_clusters)

            shash = self.model

            if 'distance' not in args:
                args['distance'] = 1

            _fit_shash, cluster_id_exactdup, matches = shash.query(**args)

            if matches.shape[0] > 0:
                # found some near duplicates
                matches_idx = np.zeros(matches.shape, dtype=np.int)
                # match the hash value to the document index
                for (i, j), value in np.ndenumerate(matches):
                    matches_idx[i, j] = shash.get_index_by_hash(value)
            else:
                matches_idx = matches
            # compute cluster_id for near duplicates
            cluster_id_dnup = _binary_linkage2clusters(matches_idx, len(_fit_shash))
            # merge near duplicates and exact duplicates clusters
            cluster_id = _merge_clusters(
                     np.concatenate((cluster_id_exactdup[:, None],
                                     cluster_id_dnup[:, None]), axis=1),
                     rename=True)
        elif self._pars['method'] == 'i-match':
            from freediscovery.near_duplicates import IMatchNearDuplicates
            if not hasattr(self, '_fit_X'):
                self._fit_X = joblib.load(str(self.fe.dsid_dir / 'features'))
            model = IMatchNearDuplicates(**args)
            model.fit(self._fit_X)
            cluster_id = model.labels_

        else:
            raise ValueError

        return cluster_id
